import os
import pandas as pd
from typing import Union

import emodpy_hiv.demographics.un_world_pop as unwp

import emodpy_hiv.plotting.xy_plot as xy_plot
import emodpy_hiv.plotting.helpers as helpers


TEST_include_dir_or_filename = True

COL_NAME_YEAR           = "Year"                         # noqa: E221
COL_NAME_NODE_ID        = " NodeId"                      # noqa: E221
COL_NAME_GENDER         = " Gender"                      # noqa: E221
COL_NAME_HAS_HIV        = " HasHIV"                      # noqa: E221
COL_NAME_RISK           = " IP_Key:Risk"                 # noqa: E221
COL_NAME_ACCESS         = " IP_Key:Accessibility"        # noqa: E221
COL_NAME_STATE          = " IP_Key:CascadeState"         # noqa: E221
COL_NAME_IS_CIRC        = " IsCircumcised"               # noqa: E221
COL_NAME_AGE            = " Age"                         # noqa: E221
COL_NAME_POP            = " Population"                  # noqa: E221
COL_NAME_INFECTED       = " Infected"                    # noqa: E221
COL_NAME_NEW_INF        = " Newly Infected"              # noqa: E221
COL_NAME_ON_ART         = " On_ART"                      # noqa: E221
COL_NAME_DIED           = " Died"                        # noqa: E221
COL_NAME_DIED_HIV       = " Died_from_HIV"               # noqa: E221
COL_NAME_TESTED_OR_ART  = " Tested Past Year or On_ART"  # noqa: E221
COL_NAME_TESTED_EVER    = " Tested Ever"                 # noqa: E221
COL_NAME_DIAGNOSED      = " Diagnosed"                   # noqa: E221
COL_NAME_NEW_TESTED_POS = " Newly Tested Positive"
COL_NAME_NEW_TESTED_NEG = " Newly Tested Negative"


def create_title(base_title: str = "",
                 node_id: int = None,
                 gender: str = None,
                 show_avg_per_run: bool = False,
                 show_fraction: bool = False,
                 show_fraction_of: bool = False,
                 fraction_of_str: str = "",
                 hiv_negative: bool = False,
                 has_age_bins: bool = False):
    """
    Use the input arguments to create a title for the plot (and filename).

    Args:
        base_title (str, optional):
            This is the core string to place in the title.  It describes the specific data being plotted.

        node_id (int, optional):
            The ID of the node for which the data is being filtered for.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        show_avg_per_run (bool, optional):
            True indicates that the data is an average over multiple runs.

        show_fraction (bool, optional):
            True indicates that the data is not true counts but a fraction
            (i.e. a count divided by another counter)

        show_fraction_of (bool, optional):
            When the denominator of the fraction can be different things,
            say population or infected, this allows you to specify the option
            that is not population.

        fraction_of_str (str, optional):
            If 'show_fraction_of' is true, then this argument an be used to include
            in the plot what the denominator is.

        hiv_negative (bool, optional):
            If True, then the title will indicate that the data is for people without HIV.

        has_age_bins (bool, optional):
            If True, then 'by Age' is added to the title.

    Returns:
        A string to be used a the top line title of the plot.
    """
    title = ""
    if show_avg_per_run and not show_fraction:
        title = "Average Per Run "
    elif not show_avg_per_run and show_fraction:
        title = title + "Fraction of "
    elif show_avg_per_run and show_fraction:
        title = "Average Fraction of "
    if show_fraction:
        if show_fraction_of:
            title = title + fraction_of_str
    if gender == "Female":
        title = title + "Females"
    elif gender == "Male":
        title = title + "Males"
    else:
        title = title + "People"

    if hiv_negative:
        title = title + " Without HIV"

    if base_title:
        title = title + " " + base_title

    if has_age_bins:
        title = title + " by Age"

    if node_id is not None:
        title = title + " - Node " + str(node_id)

    if not show_avg_per_run:
        title = title + " Per Run"

    return title


def create_y_axis_name(base_title: str = "",
                       node_id: int = None,
                       gender: str = None,
                       show_avg_per_run: bool = False,
                       show_fraction: bool = False,
                       show_fraction_of: bool = False,
                       fraction_of_str: str = "",
                       has_age_bins: bool = False):
    """
    Given the arguments, create a label for the y-axis that describes the data being plotted.

    Args:
        base_title (str, optional):
            This is the core string to place in the y-label.  It describes what is being plotted.

        node_id (int, optional):
            TBD

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        show_avg_per_run (bool, optional):
            TBD

        show_fraction (bool, optional):
            True indicates that the data is not true counts but a fraction
            (i.e. a count divided by another counter)

        show_fraction_of (bool, optional):
            When the denominator of the fraction can be different things,
            say population or infected, this allows you to specify the option
            that is not population.

        fraction_of_str (str, optional):
            If 'show_fraction_of' is true, then this argument an be used to include
            in the plot what the denominator is.

        has_age_bins (bool, optional):
            TBD

    Returns:
        A string to be used a the top line title of the plot.
    """
    y_axis_name = "Number of "
    if show_fraction:
        y_axis_name = "Fraction of "
        if show_fraction_of:
            y_axis_name = y_axis_name + fraction_of_str
    if gender == "Female":
        y_axis_name = y_axis_name + "Females"
    elif gender == "Male":
        y_axis_name = y_axis_name + "Males"
    else:
        y_axis_name = y_axis_name + "People"
    y_axis_name = y_axis_name + base_title

    return y_axis_name


def extract_population_data(filename: str,
                            node_id: int = None,
                            gender: str = None,
                            age_bin: float = None,
                            other_strat_column_name: str = None,
                            other_strat_value: Union[int, float, str] = None):
    """
    Extract population data for a specific node, gender and age.

    It is assumed that the file has ages every 5 years from 0 to 100.
    """
    df = pd.read_csv(filename)

    # -------------------------------------------
    # Verify the CSV file has the correct columns
    # -------------------------------------------
    if COL_NAME_POP not in df.columns:
        raise ValueError(f"'{COL_NAME_POP}' column does not exist in the file({filename}).")
    if other_strat_column_name and (other_strat_column_name not in df.columns):
        raise ValueError(f"'{other_strat_column_name}' column does not exist in the file({filename}).")

    if( ((other_strat_column_name is not None) and (other_strat_value is     None)) or # noqa: E201, E271, E275, W504
        ((other_strat_column_name is     None) and (other_strat_value is not None)) ): # noqa: E202, E271, E129
        raise ValueError("Both 'other_strat_column_name' and 'other_strat_value' must be specified.")

    # ---------------------------------------------------------------
    # Determine which columns to put in the pivot table and create it
    # ---------------------------------------------------------------
    pv_columns = []
    if node_id is not None:
        pv_columns.append(COL_NAME_ON_ART)
    if gender is not None:
        pv_columns.append(COL_NAME_GENDER)
    if other_strat_column_name is not None:
        pv_columns.append(other_strat_column_name)
    if age_bin is not None:
        pv_columns.append(COL_NAME_AGE)

    gender_id = 1
    if gender == "Male":
        gender_id = 0

    pv = df.pivot_table(index=COL_NAME_YEAR,
                        columns=pv_columns,
                        values=COL_NAME_POP,
                        aggfunc="sum")

    # ------------------------------------------------------------
    # Extract data from pivot table and put in dataframe.
    # If the input age_bin is 80, we need to include the data from
    # the older ages because 80 for UN World Pop is really 80+.
    # ------------------------------------------------------------
    df2 = pd.DataFrame()
    df2.index = pv.index.values
    df2[COL_NAME_POP] = 0

    list_for_tuple = []
    if node_id is not None:
        list_for_tuple.append(node_id)
    if gender is not None:
        list_for_tuple.append(gender_id)
    if other_strat_column_name is not None:
        list_for_tuple.append(other_strat_value)
    if age_bin is not None:
        age_bin_list = [age_bin]
        if age_bin == 80:
            age_bin_list = [80, 85, 90, 95, 100]
        for age in age_bin_list:
            list_for_tuple_with_age = list_for_tuple.copy()
            list_for_tuple_with_age.append(age)
            if len(list_for_tuple_with_age) == 1:
                column_tuple = list_for_tuple_with_age[0]
            else:
                column_tuple = tuple(list_for_tuple_with_age)
            df2[COL_NAME_POP] = df2[COL_NAME_POP] + pv[column_tuple]
    else:
        if len(list_for_tuple) == 1:
            column_tuple = list_for_tuple[0]
        else:
            column_tuple = tuple(list_for_tuple)
        df2[COL_NAME_POP] = pv[column_tuple]

    return df2


def extract_population_data_multiple_ages(filename: str,
                                          node_id: int = None,
                                          gender: str = None,
                                          age_bin_list: list[float] = None,
                                          filter_by_hiv_negative: bool = False,
                                          other_strat_column_name: str = None,
                                          other_strat_value: Union[int, float, str] = None,
                                          other_data_column_names: list[str] = None):
    """
    Extract population data for multiple ages for a specific node and gender.
    """
    df = pd.read_csv(filename)

    # -----------------------------------------
    # Verify the report had the correct columns
    # -----------------------------------------
    if COL_NAME_POP not in df.columns:
        raise ValueError(f"'Population' column does not exist in the file({filename}).")
    if other_strat_column_name is not None and other_strat_column_name not in df.columns:
        raise ValueError(f"'{other_strat_column_name}' column does not exist in the file({filename}).")

    if( ((other_strat_column_name is not None) and (other_strat_value is     None)) or # noqa: E201, E271, E275, W504
        ((other_strat_column_name is     None) and (other_strat_value is not None)) ): # noqa: E202, E271, E129
        raise ValueError("Both 'other_strat_column_name' and 'other_strat_value' must be specified.")

    if other_data_column_names is None:
        other_data_column_names = []

    # -----------------------------------------------------
    # Determine what data to put in to the pivot table and
    # create pivot table
    # -----------------------------------------------------
    pv_columns = []
    age_data_list = []
    if node_id is not None:
        pv_columns.append(COL_NAME_NODE_ID)
    if gender is not None:
        pv_columns.append(COL_NAME_GENDER)
    if other_strat_column_name is not None:
        pv_columns.append(other_strat_column_name)
    if age_bin_list is not None:
        pv_columns.append(COL_NAME_AGE)
        age_data_list = df[COL_NAME_AGE].unique()

    data_is_for_hiv_negative = False
    if filter_by_hiv_negative:
        if COL_NAME_HAS_HIV in df.columns:
            df = df[df[COL_NAME_HAS_HIV] == 0]
            data_is_for_hiv_negative = True

    data_columns = COL_NAME_POP
    if len(other_data_column_names) > 0:
        data_columns = []
        data_columns.append(COL_NAME_POP)
        for col_name in other_data_column_names:
            if col_name is not None and col_name not in df.columns:
                raise ValueError(f"'{col_name}' column does not exist in the file({filename}).")
            data_columns.append(col_name)

    gender_id = 1
    if gender == "Male":
        gender_id = 0

    pv = df.pivot_table(index=COL_NAME_YEAR,
                        columns=pv_columns,
                        values=data_columns,
                        aggfunc="sum")

    # -------------------------------------------------------
    # Move the data from the pivot table to a new dataframe.
    # If doing ages, we need to create labels for the columns
    # that includes the age ranges.
    # -------------------------------------------------------
    df2 = pd.DataFrame()
    df2.index = pv.index.values

    new_column_names = []
    new_column_names.append(COL_NAME_POP)
    for col_name in other_data_column_names:
        new_column_names.append(col_name)

    list_for_tuple = []
    if node_id is not None:
        list_for_tuple.append(node_id)
    if gender is not None:
        list_for_tuple.append(gender_id)
    if other_strat_column_name is not None:
        list_for_tuple.append(other_strat_value)
    if age_bin_list:
        for age_index in range(len(age_bin_list) - 1):
            age_min = age_bin_list[age_index]
            age_max = age_bin_list[age_index + 1]
            age_label = ":" + str(age_min) + " - " + str(age_max)
            for col_name in new_column_names:
                df2[col_name + age_label] = 0
            for age in age_data_list:
                if (age_min <= age) and (age < age_max):
                    for col_name in new_column_names:
                        list_for_tuple_with_age = []
                        if len(other_data_column_names) > 0:
                            list_for_tuple_with_age.append(col_name)
                        list_for_tuple_with_age.extend(list_for_tuple)
                        list_for_tuple_with_age.append(float(age))
                        if len(list_for_tuple_with_age) == 1:
                            column_tuple = list_for_tuple_with_age[0]
                        else:
                            column_tuple = tuple(list_for_tuple_with_age)
                        df2[col_name + age_label] = df2[col_name + age_label] + pv[column_tuple]
    else:
        for col_name in new_column_names:
            column_tuple = col_name
            if len(list_for_tuple) > 0:
                list_for_tuple_copy = []
                if len(other_data_column_names) > 0:
                    list_for_tuple_copy.append(col_name)
                list_for_tuple_copy.extend(list_for_tuple)
                if len(list_for_tuple_copy) == 1:
                    column_tuple = list_for_tuple_copy[0]
                else:
                    column_tuple = tuple(list_for_tuple_copy)
            df2[col_name] = pv[column_tuple]

    return df2, data_is_for_hiv_negative


def extract_population_data_by_stratification(filename: str,
                                              node_id: int = None,
                                              gender: str = None,
                                              age_bin_list: list[float] = None,
                                              start_column_name: str = None,
                                              strat_values: list[str] = None):
    """
    Extract population data such that you get the population for each stratification.
    """
    if age_bin_list is not None and len(age_bin_list) < 2:
        raise ValueError("'age_bin_list' must have at least two values.\n"
                         + "The second value is the max of the i-th bin and the min of the (i+1)-th bin.")

    df = pd.read_csv(filename)

    # --------------------------------------------
    # Verify the CSV file has the expected columns
    # --------------------------------------------
    if COL_NAME_POP not in df.columns:
        raise ValueError(f"'Population' column does not exist in the file({filename}).")
    if start_column_name not in df.columns:
        raise ValueError(f"'{start_column_name}' column does not exist in the file({filename}).")

    if strat_values is None:
        strat_values = df[start_column_name].unique()
        strat_values = [x for x in strat_values if str(x) != 'nan']

    # -----------------------------------------------------
    # Determine what data to put in to the pivot table and
    # create pivot table
    # -----------------------------------------------------
    pv_columns = []
    age_data_list = []
    pv_columns.append(start_column_name)
    if node_id is not None:
        pv_columns.append(COL_NAME_NODE_ID)
    if gender is not None:
        pv_columns.append(COL_NAME_GENDER)
    if age_bin_list is not None:
        pv_columns.append(COL_NAME_AGE)
        age_data_list = df[COL_NAME_AGE].unique()

    data_columns = COL_NAME_POP

    gender_id = 1
    if gender == "Male":
        gender_id = 0

    pv = df.pivot_table(index=COL_NAME_YEAR,
                        columns=pv_columns,
                        values=data_columns,
                        aggfunc="sum")

    # -------------------------------------------------------
    # Move the data from the pivot table to a new dataframe.
    # If doing ages, we need to create labels for the columns
    # that includes the age ranges.
    # -------------------------------------------------------
    df2 = pd.DataFrame()
    df2.index = pv.index.values

    for strat_value in strat_values:
        list_for_tuple = []
        list_for_tuple.append(strat_value)
        if node_id is not None:
            list_for_tuple.append(node_id)
        if gender is not None:
            list_for_tuple.append(gender_id)
        if age_bin_list:
            for age_index in range(len(age_bin_list) - 1):
                age_min = age_bin_list[age_index]
                age_max = age_bin_list[age_index + 1]
                age_label = ":" + str(age_min) + " - " + str(age_max)
                df2[strat_value + age_label] = 0
                for age in age_data_list:
                    if (age_min <= age) and (age < age_max):
                        list_for_tuple_with_age = list_for_tuple.copy()
                        list_for_tuple_with_age.append(float(age))
                        column_tuple = tuple(list_for_tuple_with_age)
                        df2[strat_value + age_label] = df2[strat_value + age_label] + pv[column_tuple]
        else:
            column_tuple = strat_value
            if len(list_for_tuple) > 0:
                list_for_tuple_copy = list_for_tuple.copy()
                if len(list_for_tuple_copy) == 1:
                    column_tuple = list_for_tuple_copy[0]
                else:
                    column_tuple = tuple(list_for_tuple_copy)
            df2[strat_value] = pv[column_tuple]

    return df2


def extract_population_data_by_stratification_for_dir(dir_or_filename: str,
                                                      node_id: int = None,
                                                      gender: str = None,
                                                      age_bin_list: list[float] = None,
                                                      start_column_name: str = None,
                                                      strat_values: list[str] = None,
                                                      show_avg_per_run: bool = False):
    combined_df = pd.DataFrame()

    dir_filenames = helpers.get_filenames(dir_or_filename=dir_or_filename,
                                          file_prefix="ReportHIVByAgeAndGender",
                                          file_extension=".csv")

    for fn in dir_filenames:
        print(f"Extracting data from {fn}")
        df = extract_population_data_by_stratification(filename=fn,
                                                       node_id=node_id,
                                                       gender=gender,
                                                       age_bin_list=age_bin_list,
                                                       start_column_name=start_column_name,
                                                       strat_values=strat_values)

        if len(combined_df.columns) == 0:
            combined_df.index = df.index
            for column_name in df.columns:
                name = column_name
                if not show_avg_per_run:
                    name = fn + "-" + name
                combined_df[name] = 0
        for column_name in df.columns:
            if show_avg_per_run:
                combined_df[column_name] = combined_df[column_name] + df[column_name]
            else:
                name = fn + "-" + column_name
                combined_df[name] = df[column_name]

    if show_avg_per_run:
        for column_name in combined_df.columns:
            combined_df[column_name] = combined_df[column_name] / len(dir_filenames)

    col_name_prefixs = [""]
    if not show_avg_per_run:
        col_name_prefixs = []
        for fn in dir_filenames:
            col_name_prefixs.append(fn + "-")

    return combined_df, col_name_prefixs


def create_df_for_plot_by_stratification(combined_df: pd.DataFrame,
                                         col_name_prefixs: list[str],
                                         age_bin_list: list[float] = None,
                                         show_fraction: bool = False):
    if show_fraction:
        total_df = pd.DataFrame()
        total_df.index = combined_df.index

        age_str_list = [""]
        if age_bin_list:
            for age_index in range(len(age_bin_list) - 1):
                age_min = age_bin_list[age_index]
                age_max = age_bin_list[age_index + 1]
                age_str_list.append(":" + str(age_min) + " - " + str(age_max))

        for prefix in col_name_prefixs:
            for age_str in age_str_list:
                total_label = "total-" + prefix + "-" + age_str
                total_df[total_label] = 0
                for column_name in combined_df.columns:
                    if (age_str in column_name) and (prefix in column_name):
                        total_df[total_label] = total_df[total_label] + combined_df[column_name]
                for column_name in combined_df.columns:
                    if (age_str in column_name) and (prefix in column_name):
                        combined_df[column_name] = combined_df[column_name] / total_df[total_label]

    return combined_df


def plot_population_for_dir(dir_or_filename: str,
                            unworld_pop_filename: str,
                            country: str,
                            version: str,
                            x_base_population: float = 1.0,
                            show_avg_per_run: bool = False,
                            gender: str = None,
                            age_bin: float = None,
                            img_dir: str = None):
    """
    Plot the population for the given age bin against the data in the UN World Population spreadsheet.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        unworld_pop_filename (str, required):
            The name and path to a UN World Pop Excel spreadsheet where the 'country' parameter specifies
            a country name found in the spreadsheet and the 'version' specifies the year of the data.
            These values are needed to know how to read the data in the spreadsheet.

        country (str, required):
            The name of the country found in the spreadsheet to extract the data for.

        version (str, required):
            The year associated with when the UN World Pop file was created.
            PLEASE NOTE: This year is a string.

        x_base_population (float, optional):
            The 'x_Base_Population' value (found in the config) is used to divide by the population
            numbers in the CSV file so you get numbers that match the true population.

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        age_bin (float, optional):
            If provided, the data for this specific age stratification will be plotted.
            Both the data in the report file and the UN World Pop file must have this
            stratification.  If you do not provide a value, then the population is not
            broken up by age (i.e. total population).

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if x_base_population <= 0.0:
        raise ValueError("'x_base_population' must be a value greater than zero.")

    # Get the file or files in the directory
    dir_filenames = helpers.get_filenames(dir_or_filename=dir_or_filename,
                                          file_prefix="ReportHIVByAgeAndGender",
                                          file_extension=".csv")

    # ---------------------------------------------------------------
    # Extract the data from each file and combine into one dataframe
    # ---------------------------------------------------------------
    combined_df = pd.DataFrame()
    for fn in dir_filenames:
        df = extract_population_data(fn, gender=gender, age_bin=age_bin)
        if len(combined_df.columns) == 0:
            combined_df.index = df.index
            if show_avg_per_run:
                combined_df[COL_NAME_POP] = 0
        if show_avg_per_run:
            combined_df[COL_NAME_POP] = combined_df[COL_NAME_POP] + df[COL_NAME_POP] / x_base_population
        else:
            combined_df[fn] = df[COL_NAME_POP] / x_base_population

    if show_avg_per_run:
        for column_name in combined_df.columns:
            combined_df[column_name] = combined_df[column_name] / len(dir_filenames)

    # ----------------------------------------------------------------
    # Extract the expected population data from the UN World Pop file
    # and put into dataframe.  Get the data for the years that you have
    # in the CSV file.
    # ----------------------------------------------------------------
    years = combined_df.index.values.astype(int).tolist()
    years = list(set(years))

    date_column = COL_NAME_YEAR
    if version == "2015":
        date_column = "Reference date (as of 1 July)"

    unwp_df2 = None
    title2 = dir_or_filename
    if unworld_pop_filename:
        title2 = str(unworld_pop_filename)
        unwp_df = unwp.extract_population_by_age(country=country,
                                                 version=version,
                                                 years=years,
                                                 filename=unworld_pop_filename)
        unwp_df.index = unwp_df[date_column].astype(float)
        del unwp_df[date_column]
        rename_dict = {}
        for name in unwp_df.columns:
            rename_dict[name] = int(name)
        unwp_df = unwp_df.rename(columns=rename_dict)

        unwp_df2 = pd.DataFrame()
        unwp_df2.index = unwp_df.index.astype(float)
        if age_bin is not None:
            unwp_df2["Expected Population"] = unwp_df[age_bin]
        else:
            unwp_df2["Expected Population"] = unwp_df.sum(axis=1)

    # --------------------------
    # create title and plot data
    # --------------------------
    title = ""
    if show_avg_per_run:
        title = title + "Average Per Run"
    title = title + "Population"
    if gender is not None:
        title = title + " - " + gender
    if age_bin is not None:
        if age_bin == 80:
            title = title + " - " + str(age_bin) + "+"
        else:
            title = title + " - " + str(age_bin) + "-" + str(age_bin + 5)

    if not TEST_include_dir_or_filename:
        title2 = None

    xy_plot.xy_plot(img_dir=img_dir,
                    df=combined_df,
                    expected_df=unwp_df2,
                    title_1=title,
                    title_2=title2,
                    y_axis_name="Number of People",
                    fraction_of_total=False,
                    show_legend=show_avg_per_run,
                    show_markers=show_avg_per_run,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_population_by_gender(filename: str,
                              img_dir: str = None):
    """
    For the given file, plot the population for each gender over time.

    Args
        filename (str, required):
            The name and path of the ReportHIVByAgeAndGender.csv file to extract the data from.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if not os.path.isfile(filename):
        raise ValueError(f"The filename, '{filename}' given does not appear to be a file.")

    df = pd.read_csv(filename)

    if COL_NAME_GENDER not in df.columns:
        raise ValueError(f"'{COL_NAME_GENDER}' column does not exist in the file({filename}).")

    pv = df.pivot_table(index=COL_NAME_YEAR,
                        columns=[COL_NAME_GENDER],
                        values=COL_NAME_POP,
                        aggfunc="sum")

    df2 = pd.DataFrame()
    df2.index = pv.index
    df2["Number of Men"] = pv[0]
    df2["Number of Women"] = pv[1]

    xy_plot.xy_plot(img_dir=img_dir,
                    df=df2,
                    title_1="Population by Gender",
                    title_2=None,
                    y_axis_name="Number of People",
                    fraction_of_total=False,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_population_by_ip(dir_or_filename: str,
                          exp_dir_or_filename: str = None,
                          node_id: int = None,
                          gender: str = None,
                          age_bin_list=None,
                          ip_key: str = None,
                          ip_values: list[str] = None,
                          show_avg_per_run: bool = False,
                          show_fraction: bool = False,
                          expected_values: dict = None,
                          img_dir: str = None):
    """
    For the indicated files, create a plot showing who has what value of the give IP key over time.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        exp_dir_or_filename (str, required):
            The expected or alternate directory or filename containing the ReportHIVByAgeAndGender.csv files.

        node_id (int, optional):
            The ID of the node for which the data is being filtered for.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        age_bin_list (list[float], optional):
            A list of ages, in years, where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        ip_key (str, required):
            Extract the data from the files based on this IP stratification column.
            If the files has not been stratified by this IP, you will need to re-run the
            simulations with stratification turned on.

        ip_values (list[str], optional):
            By default, this plotting tool uses all of the values for the IP, however, this allows
            you to only include the ones you are interested (i.e. a subset to the total possible
            values for the IP)

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        show_fraction (bool, optional):
            True indicates that the number of people the given IP value will be divided by
            the sum of all the people with all of the selected IP values.  For example, if the
            IP key were Risk and you selected to only plot LOW and MEDIUM, then the total number
            of people with LOW will be divided by the total number of people with either LOW or MEDIUM.

        expected_values (dict, optional):
            If the user provides this dictionary, the constant expected value of each IP will be
            plotted.  This should be a dictionary with an IP value as the key and a constant expected value.
            There should be an IP value for each IP value plotted.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """

    if (exp_dir_or_filename is not None) and (expected_values is not None):
        raise ValueError("You cannot specify both 'exp_dir_or_filename' and 'expected_values'. "
                         + "Please only specify one of these.")

    if ip_key is None:
        raise ValueError("IP Key must be specified.")
    ip_col_name = " IP_Key:" + ip_key

    combined_df, col_name_prefixs = extract_population_data_by_stratification_for_dir(dir_or_filename=dir_or_filename,
                                                                                      node_id=node_id,
                                                                                      gender=gender,
                                                                                      age_bin_list=age_bin_list,
                                                                                      start_column_name=ip_col_name,
                                                                                      strat_values=ip_values,
                                                                                      show_avg_per_run=show_avg_per_run)

    combined_df = create_df_for_plot_by_stratification(combined_df=combined_df,
                                                       col_name_prefixs=col_name_prefixs,
                                                       age_bin_list=age_bin_list,
                                                       show_fraction=show_fraction)

    expected_df = None
    if expected_values:
        expected_df = pd.DataFrame()
        expected_df.index = combined_df.index
        for value in ip_values:
            expected_df["Expected: " + value] = expected_values[value]
    elif exp_dir_or_filename is not None:
        expected_df, exp_col_name_prefixs = extract_population_data_by_stratification_for_dir(dir_or_filename=exp_dir_or_filename,
                                                                                              node_id=node_id,
                                                                                              gender=gender,
                                                                                              age_bin_list=age_bin_list,
                                                                                              start_column_name=ip_col_name,
                                                                                              strat_values=ip_values,
                                                                                              show_avg_per_run=show_avg_per_run)
        expected_df = create_df_for_plot_by_stratification(combined_df=expected_df,
                                                           col_name_prefixs=exp_col_name_prefixs,
                                                           age_bin_list=age_bin_list,
                                                           show_fraction=show_fraction)

    base_title = "with " + ip_key + "=X"
    title = create_title(base_title=base_title,
                         node_id=node_id,
                         gender=gender,
                         show_avg_per_run=show_avg_per_run,
                         show_fraction=show_fraction,
                         show_fraction_of=False,
                         fraction_of_str=None,
                         has_age_bins=(age_bin_list is not None))
    y_axis_name = create_y_axis_name(base_title=" " + base_title,
                                     node_id=node_id,
                                     gender=gender,
                                     show_avg_per_run=show_avg_per_run,
                                     show_fraction=show_fraction,
                                     show_fraction_of=False,
                                     fraction_of_str=None,
                                     has_age_bins=(age_bin_list is not None))

    title2 = None
    if TEST_include_dir_or_filename:
        if exp_dir_or_filename is not None:
            title2 = "color= " + dir_or_filename
            title2 = title2 + "\nblack = " + exp_dir_or_filename
        else:
            title2 = dir_or_filename

    xy_plot.xy_plot(img_dir=img_dir,
                    df=combined_df,
                    expected_df=expected_df,
                    title_1=title,
                    title_2=title2,
                    y_axis_name=y_axis_name,
                    show_legend=show_avg_per_run,
                    fraction_of_total=False,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_columns(filename: str,
                 title: str,
                 y_axis_name: str,
                 column_names: list[str],
                 fraction_of_population: bool = False,
                 img_dir: str = None):
    """
    For a given file, plot the indicated columns versus time (i.e. Year).

    Args:
        filename (str, required):
            The name and path of the ReportHIVByAgeAndGender.csv file to extract the data from.

        title (str, required):
            The title to put at the top of the plot.

        y_axis_name (str, required):
            The name to label the y-axis on the plot.

        column_names (list[str], required):
            The list of column names to plot the data for.  The report has a space before each
            column name.  Please be sure to include it.

        fraction_of_population (bool, optional):
            If True, divide the count each column by the population for the same stratification.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if not os.path.isfile(filename):
        raise ValueError(f"The filename, '{filename}' given does not appear to be a file.")

    df = pd.read_csv(filename)

    for name in column_names:
        if name not in df.columns:
            raise ValueError(f"'{name}' column does not exist in the file({filename}).")

    if fraction_of_population:
        column_names.append(COL_NAME_POP)

    pv = df.pivot_table(index=COL_NAME_YEAR,
                        columns=[],
                        values=column_names,
                        aggfunc="sum")

    df2 = pd.DataFrame()
    df2.index = pv.index
    for value in column_names:
        if fraction_of_population and (value != COL_NAME_POP):
            df2[value] = pv[value] / pv[COL_NAME_POP]
        elif not fraction_of_population:
            df2[value] = pv[value]

    xy_plot.xy_plot(img_dir=img_dir,
                    df=df2,
                    title_1=title,
                    title_2=None,
                    y_axis_name=y_axis_name,
                    fraction_of_total=False,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_circumcision_by_age(filename: str,
                             age_bin_list: list[float],
                             fraction_of_total: bool = False,
                             img_dir: str = None):
    """
    For a single file, plot the number of men who are circumcised by age.

    Args:
        filename (str, required):
            The name and path of the ReportHIVByAgeAndGender.csv to extract the data from.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        fraction_of_total (bool, optional):
            If True, the number of men who are circumcised will be divided by the total number
            of men with that stratification.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if not os.path.isfile(filename):
        raise ValueError(f"The filename, '{filename}' given does not appear to be a file.")

    df = pd.read_csv(filename)

    if COL_NAME_GENDER not in df.columns:
        raise ValueError(f"'{COL_NAME_GENDER}' column does not exist in the file({filename}).")
    if COL_NAME_AGE not in df.columns:
        raise ValueError(f"'{COL_NAME_AGE}' column does not exist in the file({filename}).")

    ages = df[COL_NAME_AGE].unique()

    y_axis_name = "Number of Circumcised Men"
    if fraction_of_total:
        y_axis_name = "Fraction of Men Circumcised by Age"

    pv = df.pivot_table(index=COL_NAME_YEAR,
                        columns=[COL_NAME_GENDER, COL_NAME_AGE, COL_NAME_IS_CIRC],
                        values=COL_NAME_POP,
                        aggfunc="sum")

    df2 = pd.DataFrame()
    df2.index = pv.index

    gender_id = 0
    is_circumcised = 1
    for index in range(len(age_bin_list) - 1):
        label = f"Men Circumcised: {age_bin_list[index]}-{age_bin_list[index + 1]}"
        pv[label] = 0
        pv["total"] = 0
        for age in ages:
            if (age_bin_list[index] <= age) and (age < age_bin_list[index + 1]):
                pv["total"] = pv["total"] + pv[(gender_id, age, 0)] + pv[(gender_id, age, 1)]
                pv[label] = pv[label] + pv[(gender_id, age, is_circumcised)]
        df2[label] = pv[label]
        if fraction_of_total:
            df2[label] = df2[label] / pv["total"]

    xy_plot.xy_plot(img_dir=img_dir,
                    df=df2,
                    title_1="Circumcised Men by Age",
                    title_2=None,
                    y_axis_name=y_axis_name,
                    fraction_of_total=False,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def extract_population_data_multiple_ages_for_dir(dir_or_filename: str,
                                                  node_id: int = None,
                                                  gender: str = None,
                                                  age_bin_list: list[float] = None,
                                                  show_avg_per_run: bool = False,
                                                  filter_by_hiv_negative: bool = False,
                                                  other_strat_column_name: str = None,
                                                  other_strat_value_a: Union[int, float, str] = None,
                                                  other_strat_value_b: Union[int, float, str] = None,
                                                  other_data_column_names: list[str] = None):
    combined_df = pd.DataFrame()

    data_is_for_hiv_negative = False
    dir_filenames = helpers.get_filenames(dir_or_filename=dir_or_filename,
                                          file_prefix="ReportHIVByAgeAndGender",
                                          file_extension=".csv")
    for fn in dir_filenames:
        print("extracting data from " + fn)
        df_a, hiv_neg_a = extract_population_data_multiple_ages(fn,
                                                                node_id=node_id,
                                                                gender=gender,
                                                                age_bin_list=age_bin_list,
                                                                filter_by_hiv_negative=filter_by_hiv_negative,
                                                                other_strat_column_name=other_strat_column_name,
                                                                other_strat_value=other_strat_value_a,
                                                                other_data_column_names=other_data_column_names)
        df_b = None
        hiv_neg_b = None
        if other_strat_column_name is not None and other_strat_value_b is not None:
            df_b, hiv_neg_b = extract_population_data_multiple_ages(fn,
                                                                    node_id=node_id,
                                                                    gender=gender,
                                                                    age_bin_list=age_bin_list,
                                                                    filter_by_hiv_negative=filter_by_hiv_negative,
                                                                    other_strat_column_name=other_strat_column_name,
                                                                    other_strat_value=other_strat_value_b,
                                                                    other_data_column_names=other_data_column_names)

        if hiv_neg_b is not None and hiv_neg_a != hiv_neg_b:
            raise ValueError("The two dataframes for the two stratifications do not have the same HIV negative status. "
                             + "This is likely because you are trying to extract data for a stratification that does not exist in the file.")
        data_is_for_hiv_negative = hiv_neg_a

        if len(combined_df.columns) == 0:
            combined_df.index = df_a.index
            for column_name in df_a.columns:
                name = column_name
                if not show_avg_per_run:
                    name = fn + "-" + name
                combined_df[name] = 0
        for column_name in df_a.columns:
            if show_avg_per_run:
                if df_b is not None:
                    combined_df[column_name] = combined_df[column_name] + df_a[column_name] / (df_a[column_name] + df_b[column_name])
                else:
                    combined_df[column_name] = combined_df[column_name] + df_a[column_name]
            else:
                name = fn + "-" + column_name
                if df_b is not None:
                    combined_df[name] = df_a[column_name] / (df_a[column_name] + df_b[column_name])
                else:
                    combined_df[name] = df_a[column_name]

    if show_avg_per_run:
        for column_name in combined_df.columns:
            combined_df[column_name] = combined_df[column_name] / len(dir_filenames)

    col_name_prefixs = [""]
    if not show_avg_per_run:
        col_name_prefixs = []
        for fn in dir_filenames:
            col_name_prefixs.append(fn + "-")

    return combined_df, col_name_prefixs, data_is_for_hiv_negative


def create_df_for_plot_by_age(combined_df: pd.DataFrame,
                              col_name_prefixs: list[str],
                              main_column_name: str,
                              gender: str,
                              age_bins: list[float],
                              show_fraction: bool,
                              fraction_of: bool,
                              fraction_of_column_name: str):
    df2 = pd.DataFrame()
    df2.index = combined_df.index

    if show_fraction:
        for prefix in col_name_prefixs:
            fraction_label = prefix + main_column_name + "/Population"
            if fraction_of:
                fraction_label = prefix + main_column_name + "/" + fraction_of_column_name[1:]
            if age_bins:
                for age_index in range(len(age_bins) - 1):
                    age_min = age_bins[age_index    ]  # noqa: E202
                    age_max = age_bins[age_index + 1]
                    main_label         = prefix + main_column_name        + ":" + str(age_min) + " - " + str(age_max) # noqa: E221
                    pop_label          = prefix + COL_NAME_POP            + ":" + str(age_min) + " - " + str(age_max) # noqa: E221
                    other_label        = prefix + fraction_of_column_name + ":" + str(age_min) + " - " + str(age_max) # noqa: E221
                    fraction_label_age = fraction_label                   + ":" + str(age_min) + " - " + str(age_max) # noqa: E221
                    if fraction_of:
                        df2[fraction_label_age] = combined_df[main_label] / combined_df[other_label]
                    else:
                        df2[fraction_label_age] = combined_df[main_label] / combined_df[pop_label]
                    df2[fraction_label_age] = df2[fraction_label_age].fillna(0)
            else:
                if fraction_of:
                    df2[fraction_label] = combined_df[prefix + main_column_name] / combined_df[prefix + fraction_of_column_name]
                else:
                    df2[fraction_label] = combined_df[prefix + main_column_name] / combined_df[prefix + COL_NAME_POP]
                df2[fraction_label] = df2[fraction_label].fillna(0)
    else:
        new_column_names = {}
        for column_name in combined_df.columns:
            if main_column_name not in column_name:
                del combined_df[column_name]
            else:
                label = column_name
                if gender:
                    label = gender + column_name
                new_column_names[column_name] = label
        combined_df = combined_df.rename(columns=new_column_names)
        df2 = combined_df.copy()

    return df2


def base_plot_by_age(base_title: str,
                     main_column_name: str,
                     dir_or_filename: str,
                     exp_dir_or_filename: str = None,
                     node_id: int = None,
                     age_bins: list[float] = None,
                     gender: str = None,
                     filter_by_hiv_negative: bool = False,
                     other_strat_column_name: str = None,
                     other_strat_value_a: Union[int, float, str] = None,
                     other_strat_value_b: Union[int, float, str] = None,
                     show_avg_per_run: bool = False,
                     show_fraction: bool = False,
                     fraction_of: bool = False,
                     fraction_of_column_name: str = None,
                     fraction_of_str: str = None,
                     img_dir: str = None):

    other_data_column_names = []
    if main_column_name != COL_NAME_POP:
        other_data_column_names.append(main_column_name)
    if show_fraction and fraction_of:
        other_data_column_names.append(fraction_of_column_name)

    combined_df, col_name_prefixs, hiv_neg = extract_population_data_multiple_ages_for_dir(dir_or_filename=dir_or_filename,
                                                                                           node_id=node_id,
                                                                                           gender=gender,
                                                                                           age_bin_list=age_bins,
                                                                                           show_avg_per_run=show_avg_per_run,
                                                                                           filter_by_hiv_negative=filter_by_hiv_negative,
                                                                                           other_strat_column_name=other_strat_column_name,
                                                                                           other_strat_value_a=other_strat_value_a,
                                                                                           other_strat_value_b=other_strat_value_b,
                                                                                           other_data_column_names=other_data_column_names)

    df2 = create_df_for_plot_by_age(combined_df=combined_df,
                                    col_name_prefixs=col_name_prefixs,
                                    main_column_name=main_column_name,
                                    gender=gender,
                                    age_bins=age_bins,
                                    show_fraction=show_fraction,
                                    fraction_of=fraction_of,
                                    fraction_of_column_name=fraction_of_column_name)

    exp_df2 = None
    exp_hiv_neg = None
    if exp_dir_or_filename:
        exp_combined_df, exp_col_name_prefixs, exp_hiv_neg = extract_population_data_multiple_ages_for_dir(dir_or_filename=exp_dir_or_filename,
                                                                                                           node_id=node_id,
                                                                                                           gender=gender,
                                                                                                           age_bin_list=age_bins,
                                                                                                           show_avg_per_run=show_avg_per_run,
                                                                                                           filter_by_hiv_negative=filter_by_hiv_negative,
                                                                                                           other_strat_column_name=other_strat_column_name,
                                                                                                           other_strat_value_a=other_strat_value_a,
                                                                                                           other_strat_value_b=other_strat_value_b,
                                                                                                           other_data_column_names=other_data_column_names)
        exp_df2 = create_df_for_plot_by_age(combined_df=exp_combined_df,
                                            col_name_prefixs=exp_col_name_prefixs,
                                            main_column_name=main_column_name,
                                            gender=gender,
                                            age_bins=age_bins,
                                            show_fraction=show_fraction,
                                            fraction_of=fraction_of,
                                            fraction_of_column_name=fraction_of_column_name)

    if exp_hiv_neg is not None and hiv_neg != exp_hiv_neg:
        raise ValueError("The two sets of data do not each have the HasHIV column.\n"
                         + "The files in {dir_or_filename} have HasHIV={hiv_neg}.\n"
                         + "The files in {exp_dir_or_filename} have HasHIV={exp_hiv_neg}.\n"
                         + "Please check the files and make sure they are correct.")

    title = create_title(base_title=base_title,
                         node_id=node_id,
                         gender=gender,
                         show_avg_per_run=show_avg_per_run,
                         show_fraction=show_fraction,
                         show_fraction_of=fraction_of,
                         fraction_of_str=fraction_of_str,
                         hiv_negative=hiv_neg,
                         has_age_bins=(age_bins is not None))
    y_axis_name = create_y_axis_name(base_title=" " + base_title,
                                     node_id=node_id,
                                     gender=gender,
                                     show_avg_per_run=show_avg_per_run,
                                     show_fraction=show_fraction,
                                     show_fraction_of=fraction_of,
                                     fraction_of_str=fraction_of_str,
                                     has_age_bins=(age_bins is not None))

    title2 = None
    if TEST_include_dir_or_filename:
        if exp_df2 is not None:
            title2 = "color=" + dir_or_filename
            title2 = title2 + "\nblack=" + exp_dir_or_filename
        else:
            title2 = dir_or_filename

    xy_plot.xy_plot(img_dir=img_dir,
                    df=df2,
                    expected_df=exp_df2,
                    title_1=title,
                    title_2=title2,
                    y_axis_name=y_axis_name,
                    show_legend=show_avg_per_run,
                    fraction_of_total=False,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_onART_by_age(dir_or_filename: str,
                      exp_dir_or_filename: str = None,
                      node_id: int = None,
                      gender: str = None,
                      age_bin_list: list[float] = None,
                      show_avg_per_run: bool = False,
                      show_fraction: bool = False,
                      fraction_of_infected: bool = False,
                      img_dir: str = None):
    """
    Create a plot showing information about the people on ART.  You can show the fraction of the
    population or the fraction of the infected population.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        exp_dir_or_filename (str, required):
            The expected or alternate directory or filename containing the ReportHIVByAgeAndGender.csv files.

        node_id (int, optional):
            The ID of the node for which the data is being filtered for.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        show_fraction (bool, optional):
            True indicates that the number of people on ART will be divided by either the number
            of people in the population or the number of infected people.  It depends on the
            'fraction_of_infected' parameter.

        fraction_of_infected (bool, optional):
            If 'show_fraction' is True, then this parameter determines what the divisor is when
            creating the fraction.  If it is True, it will divide the number of people on ART
            by the number of infected people.  If it is False, the divisor will the total population
            with that stratification.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    base_plot_by_age(base_title="On ART",
                     main_column_name=COL_NAME_ON_ART,
                     dir_or_filename=dir_or_filename,
                     exp_dir_or_filename=exp_dir_or_filename,
                     node_id=node_id,
                     gender=gender,
                     age_bins=age_bin_list,
                     show_avg_per_run=show_avg_per_run,
                     show_fraction=show_fraction,
                     fraction_of=fraction_of_infected,
                     fraction_of_column_name=COL_NAME_INFECTED,
                     fraction_of_str="Infected ",
                     img_dir=img_dir)


def plot_population_by_age(dir_or_filename: str,
                           exp_dir_or_filename: str = None,
                           node_id: int = None,
                           gender: str = None,
                           age_bin_list: list[float] = None,
                           show_avg_per_run: bool = False,
                           img_dir: str = None):
    """
    Create a plot showing the population over time.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        exp_dir_or_filename (str, required):
            The expected or alternate directory or filename containing the ReportHIVByAgeAndGender.csv files.

        node_id (int, optional):
            The ID of the node for which the data is being filtered for.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if age_bin_list is not None and len(age_bin_list) == 0:
        raise ValueError("The 'age_bin_list' parameter must be a list of ages in years where the population will be counted for each bin. "
                         + "If you do not want to use age bins, please set it to None.")

    base_plot_by_age(base_title="",
                     main_column_name=COL_NAME_POP,
                     dir_or_filename=dir_or_filename,
                     exp_dir_or_filename=exp_dir_or_filename,
                     node_id=node_id,
                     gender=gender,
                     age_bins=age_bin_list,
                     show_avg_per_run=show_avg_per_run,
                     show_fraction=False,
                     fraction_of=False,
                     fraction_of_column_name=None,
                     fraction_of_str=None,
                     img_dir=img_dir)


def plot_vmmc_by_age(dir_or_filename: str,
                     exp_dir_or_filename: str = None,
                     node_id: int = None,
                     age_bin_list: list[float] = None,
                     show_avg_per_run: bool = False,
                     img_dir: str = None):
    """
    Create a plot showing information about the men who are circumcised.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        exp_dir_or_filename (str, required):
            The expected or alternate directory or filename containing the ReportHIVByAgeAndGender.csv files.

        node_id (int, optional):
            The ID of the node for which the data is being filtered for.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    base_plot_by_age(base_title="VMMC",
                     main_column_name=COL_NAME_POP,
                     dir_or_filename=dir_or_filename,
                     exp_dir_or_filename=exp_dir_or_filename,
                     node_id=node_id,
                     gender="Male",
                     age_bins=age_bin_list,
                     show_avg_per_run=show_avg_per_run,
                     show_fraction=False,
                     fraction_of=False,
                     fraction_of_column_name="",
                     fraction_of_str=None,
                     filter_by_hiv_negative=True,
                     other_strat_column_name=COL_NAME_IS_CIRC,
                     other_strat_value_a=1,  # Only show circumcised
                     other_strat_value_b=0,  # Not circumcised
                     img_dir=img_dir)


def plot_population_by_age_vs_unworld_pop(filename: str,
                                          unworld_pop_filename: str,
                                          age_bin_list: list[float],
                                          country: str,
                                          version: str,
                                          x_base_population: float = 1.0,
                                          img_dir=None):
    """
    For a single file, plot the actual population for a given age bin versus what was expected
    in the given UN World Population file.  If you have define three age bins, there will be
    six curves -  an actual and an expected for each bin.

    Args:
        filename (str, required):
            The name and path of the ReportHIVByAgeAndGender.csv file to plot the data from.

        unworld_pop_filename (str, required):
            The name and path to a UN World Pop Excel spreadsheet where the 'country' parameter specifies
            a country name found in the spreadsheet and the 'version' specifies the year of the data.
            These values are needed to know how to read the data in the spreadsheet.

        age_bin_list (list[float], required):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        country (str, required):
            The name of the country found in the spreadsheet to extract the data for.

        version (str, required):
            The year associated with when the UN World Pop file was created.
            PLEASE NOTE: This year is a string.

        x_base_population (float, optional):
            The 'x_Base_Population' value (found in the config) is used to divide by the population
            numbers in the CSV file so you get numbers that match the true population.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if not os.path.isfile(filename):
        raise ValueError(f"The filename, '{filename}' given does not appear to be a file.")

    df = pd.read_csv(filename)

    if COL_NAME_GENDER not in df.columns:
        raise ValueError(f"'{COL_NAME_GENDER}' column does not exist in the file({filename}).")
    if COL_NAME_AGE not in df.columns:
        raise ValueError(f"'{COL_NAME_AGE}' column does not exist in the file({filename}).")
    if COL_NAME_POP not in df.columns:
        raise ValueError(f"'{COL_NAME_POP}' column does not exist in the file({filename}).")

    ages = df[COL_NAME_AGE].unique()

    pv = df.pivot_table(index=COL_NAME_YEAR,
                        columns=[COL_NAME_AGE],
                        values=COL_NAME_POP,
                        aggfunc="sum")

    df2 = pd.DataFrame()
    df2.index = pv.index

    for index in range(len(age_bin_list) - 1):
        label = f"Population: {age_bin_list[index]}-{age_bin_list[index + 1]}"
        pv[label] = 0
        for age in ages:
            if (age_bin_list[index] <= age) and (age < age_bin_list[index + 1]):
                pv[label] = pv[label] + pv[(age)]
        df2[label] = pv[label] / x_base_population

    years = df2.index.values.astype(int).tolist()
    years = list(set(years))

    unwp_df = unwp.extract_population_by_age(country=country,
                                             version=version,
                                             years=years,
                                             filename=unworld_pop_filename)
    unwp_df.index = unwp_df[COL_NAME_YEAR]

    unwp_df2 = pd.DataFrame()
    unwp_df2.index = unwp_df.index

    for index in range(len(age_bin_list) - 1):
        name_old = str(age_bin_list[index])
        name_new = f"Expected Population: {age_bin_list[index]}-{age_bin_list[index + 1]}"
        unwp_df2[name_new] = unwp_df[name_old]

    unwp_df2.index = unwp_df2.index.astype(float)

    xy_plot.xy_plot(img_dir=img_dir,
                    df=df2,
                    expected_df=unwp_df2,
                    title_1="Population by Age",
                    title_2=f"UN World Population - {country} - {version}",
                    y_axis_name="Number of People",
                    fraction_of_total=False,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_risk(dir_or_filename: str,
              starting_expected_values: dict = None,
              expected_value_for_high_per_node: list[float] = None,
              gender: str = None,
              age_bin_list: list[float] = None,
              show_avg_per_run: bool = False,
              show_fraction: bool = False,
              img_dir: str = None):
    """
    Create one plot for each node in 'expected_value_for_high_per_node' showing the risk
    values for the population versus what is expected.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        starting_expected_values (dict, optional):
            The starting three values of how risk distributed to the population.  These
            should be the same values that you have in the demographics.  The order is
            LOW, MEDIUM, and HIGH.  Typically, one might have 0.85 for LOW, 0.15 for MEDIUM,
            and 0.0 for HIGH because people get set to HIGH in the campaign's CSW logic.

        expected_value_for_high_per_node (list[float, optional]):
            A list of expected fraction of the population to have Risk = HIGH for a given node.
            The node ID of each value is expected to be the index of the position plus 1.
            The starting values for LOW and MEDIUM are adjusted for this HIGH value.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        show_fraction (bool, optional):
            True indicates that for each stratification the number of people with a given
            risk value is divided by the total number of people in that stratification.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """

    if ((starting_expected_values is not None) and (expected_value_for_high_per_node is     None) or    # noqa: E271, W504
        (starting_expected_values is     None) and (expected_value_for_high_per_node is not None)):     # noqa: E225, E271, E129
        raise ValueError("'starting_expected_values' and 'expected_value_for_high_per_node'" +          # noqa: W504
                         " need to be either both None or both defined.")

    if (starting_expected_values is not None) and (starting_expected_values["HIGH"] != 0.0):
        raise ValueError("Expected 'starting_expected_values['HIGH'] to be zero.")

    if (starting_expected_values is not None) and (expected_value_for_high_per_node is not None):
        node_id = 0
        for high_value in expected_value_for_high_per_node:
            node_id = node_id + 1

            expected_values = None
            if (starting_expected_values is not None) and (expected_value_for_high_per_node is not None):
                expected_values = {}
                expected_values["LOW"   ] = starting_expected_values["LOW"   ] * (1.0 - high_value) # noqa: E202
                expected_values["MEDIUM"] = starting_expected_values["MEDIUM"] * (1.0 - high_value)
                expected_values["HIGH"  ] = high_value # noqa: E202

            plot_population_by_ip(dir_or_filename=dir_or_filename,
                                  node_id=node_id,
                                  gender=gender,
                                  age_bin_list=age_bin_list,
                                  ip_key="Risk",
                                  ip_values=["LOW", "MEDIUM", "HIGH"],
                                  show_avg_per_run=show_avg_per_run,
                                  show_fraction=show_fraction,
                                  expected_values=expected_values,
                                  img_dir=img_dir)
    else:
        plot_population_by_ip(dir_or_filename=dir_or_filename,
                              node_id=None,
                              gender=gender,
                              age_bin_list=age_bin_list,
                              ip_key="Risk",
                              ip_values=["LOW", "MEDIUM", "HIGH"],
                              show_avg_per_run=show_avg_per_run,
                              show_fraction=show_fraction,
                              expected_values=None,
                              img_dir=img_dir)


def plot_vmmc_for_dir(dir_or_filename: str,
                      node_id: int = None,
                      age_bin_list: list[float] = None,
                      show_expected: bool = False,
                      show_avg_per_run: bool = False,
                      img_dir: str = None):
    """
    Create a plot showing what fraction of men are circumcised over time versus the
    expected number of circumcised men.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        node_id (int, optional):
            The ID of the node for which the data is being filtered for.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        show_expected (bool, optional):
            If true, plot the expected fraction of circumcisions.

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if node_id is None and show_expected:
        raise ValueError("You need to specify 'node_id' if you want to compare actual data against the expected values.")

    dir_filenames = helpers.get_filenames(dir_or_filename=dir_or_filename,
                                          file_prefix="ReportHIVByAgeAndGender",
                                          file_extension=".csv")

    combined_df = pd.DataFrame()
    for fn in dir_filenames:
        df_circ, not_used = extract_population_data_multiple_ages(fn,
                                                                  node_id=node_id,
                                                                  gender="Male",
                                                                  age_bin_list=age_bin_list,
                                                                  other_strat_column_name=COL_NAME_IS_CIRC,
                                                                  other_strat_value=1)
        df_un_circ, not_used = extract_population_data_multiple_ages(fn,
                                                                     node_id=node_id,
                                                                     gender="Male",
                                                                     age_bin_list=age_bin_list,
                                                                     other_strat_column_name=COL_NAME_IS_CIRC,
                                                                     other_strat_value=0)
        if len(combined_df.columns) == 0:
            combined_df.index = df_circ.index
            for column_name in df_circ.columns:
                name = column_name
                if not show_avg_per_run:
                    name = fn + "-" + name
                combined_df[name] = 0
        for column_name in df_circ.columns:
            if show_avg_per_run:
                combined_df[column_name] = combined_df[column_name] + df_circ[column_name] / (df_circ[column_name] + df_un_circ[column_name])
            else:
                name = fn + "-" + column_name
                combined_df[name] = df_circ[column_name] / (df_circ[column_name] + df_un_circ[column_name])

    if show_avg_per_run:
        for column_name in combined_df.columns:
            combined_df[column_name] = combined_df[column_name] / len(dir_filenames)

    traditional_coverage_per_node = [0.054978651, 0.139462861, 0.028676043, 0.091349358, 0.123187070,
                                     0.039308099, 0.727917322, 0.041105263, 0.044388102, 0.398239794]
    rtec_x1 = 2016
    rtec_x2 = 2021
    rtec_y1 = 0.54
    rtec_y2 = 0.9

    expected_df = None
    if show_expected:
        expected_df = pd.DataFrame()
        expected_df.index = combined_df.index
        expected_df["Expected Traditional"] = traditional_coverage_per_node[node_id - 1]
        years = combined_df.index.values.astype(float).tolist()
        rtec = []
        for year in years:
            if year < rtec_x1:
                rtec.append(0)
            elif year > rtec_x2:
                rtec.append(rtec_y2)
            else:
                rtec.append(rtec_y1 + (year - rtec_x1) * (rtec_y2 - rtec_y1) / (rtec_x2 - rtec_x1))
        expected_df["Expected Plus Medical"] = rtec

    title = "Circumcised Men"
    if show_avg_per_run:
        title = title + " - Average"
    if node_id is not None:
        title = title + " - Node " + str(node_id)

    title2 = None
    if TEST_include_dir_or_filename:
        title2 = dir_or_filename

    xy_plot.xy_plot(img_dir=img_dir,
                    df=combined_df,
                    expected_df=expected_df,
                    title_1=title,
                    title_2=title2,
                    y_axis_name="Fraction of Circumcised Men",
                    fraction_of_total=False,
                    show_legend=show_avg_per_run,
                    show_markers=show_avg_per_run,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_prevalence_for_dir(dir_or_filename: str,
                            exp_dir_or_filename: str = None,
                            node_id: int = None,
                            gender: str = None,
                            age_bin_list: list[float] = None,
                            show_avg_per_run: bool = False,
                            show_fraction: bool = False,
                            img_dir: str = None):
    """
    Create a plot showing who is infected with HIV.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        exp_dir_or_filename (str, required):
            The expected or alternate directory or filename containing the ReportHIVByAgeAndGender.csv files.

        node_id (int, optional):
            The ID of the node for which the data is being filtered for.

        gender (str, optional):
            The string (Male or Female) for the gender that data is being filtered for.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        show_fraction (bool, optional):
            True indicates that the number of infected people should be divided by the total
            number of people in that stratification.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    base_plot_by_age(base_title="Infected",
                     main_column_name=COL_NAME_INFECTED,
                     dir_or_filename=dir_or_filename,
                     exp_dir_or_filename=exp_dir_or_filename,
                     node_id=node_id,
                     age_bins=age_bin_list,
                     gender=gender,
                     show_avg_per_run=show_avg_per_run,
                     show_fraction=show_fraction,
                     fraction_of=False,
                     fraction_of_column_name="",
                     fraction_of_str="",
                     img_dir=img_dir)


def plot_risk_zambia(dir_or_filename: str,
                     age_bin_list: list[float] = None,
                     show_avg_per_run: bool = False,
                     show_fraction: bool = False,
                     show_expected: bool = False,
                     img_dir: str = None):
    """
    Create multiple risk value plots where each plot is for a specific node and gender.
    The plot can show the count or fraction of the group that has one of the three
    risk values.  The plot can also show the expected values for specific node and gender
    for Zambia.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the ReportHIVByAgeAndGender.csv files.

        age_bin_list (list[float], optional):
            A list of ages in years where the population with risk value will be counted
            for each bin.  For example, if you enter [10, 25, 30, 55], there will be three
            age bins with the following ranges: [10-25), [25-30), [30-55)

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            people with the given risk type at a given time step between the files.
            Default is False.

        show_fraction (bool, optional):
            True indicates that the data is not true counts but a fraction
            (i.e. a count divided by another counter)

        show_expected (bool, optional):
            True indicates that you want to see the expected fractions of the population that
            have the different risk values PER NODE.  False will be data for all nodes.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    starting_expected_values = {
        "LOW": 0.85,
        "MEDIUM": 0.15,
        "HIGH": 0.0
    }
    node_to_high = []
    gender_list = ["Female", "Male"]
    for gender in gender_list:
        if gender == "Female":
            # node_to_high = [0.125, 0.125, 0.125, 0.0500, 0.0500, 0.0500, 0.125, 0.0500, 0.125, 0.0500]
            node_to_high = [0.125, 0.125, 0.125, 0.0500]
        else:
            # node_to_high = [0.195, 0.195, 0.195, 0.0781, 0.0781, 0.0781, 0.195, 0.0781, 0.195, 0.0781]
            node_to_high = [0.195, 0.195, 0.195, 0.0781]

        if not show_expected:
            starting_expected_values = None
            node_to_high = None
        plot_risk(dir_or_filename=dir_or_filename,
                  starting_expected_values=starting_expected_values,
                  expected_value_for_high_per_node=node_to_high,
                  gender=gender,
                  age_bin_list=age_bin_list,
                  show_avg_per_run=show_avg_per_run,
                  show_fraction=show_fraction,
                  img_dir=img_dir)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument('dir_or_filename', type=str, nargs=1, help='A directory with ReportHivByAgeAndGender.csv files or a single file.')
    parser.add_argument('exp_dir_or_filename', type=str, default=None, nargs='?', help='A directory with ReportHivByAgeAndGender.csv files or a single file to compare to.')
    parser.add_argument('-p', '--plot', default='population', help='Options: population, prevalence, risk, vmmc, art, state, column, summary')
    parser.add_argument('-o', '--output', default=None, help='If provided, a directory will be created and images saved to the folder.  If not provided, it opens windows.')
    parser.add_argument('-n', '--node_id', type=int, default=None, help='Plot the data for a specific node.')
    parser.add_argument('-m', '--mean', help='Gives the average/mean of each run at that time point.', action='store_true')
    parser.add_argument('-a', '--ages', help='Show the data stratified by age.', action='store_true')
    parser.add_argument('-x', '--expected', help='Show the expected data.', action='store_true')
    parser.add_argument('-c', '--column_name', default='Died', help='Only used with plot=column.  The name of a non-stratification column from the report.  Do not include the space before the name.')

    args = parser.parse_args()

    dir_or_filename = args.dir_or_filename[0]
    exp_dir_or_filename = None
    if args.exp_dir_or_filename is not None:
        exp_dir_or_filename = args.exp_dir_or_filename

    node_id = args.node_id

    plot_list = []
    if args.plot == "summary":
        plot_list = ["population", "prevalence", "vmmc", "art", "state"]
    else:
        plot_list = [args.plot]

    for plot in plot_list:
        if plot == "population":
            if args.expected:
                print("===============================================")
                print("Prevalence plot does not support expected data.")
                print("===============================================")

            age_bin_list = None
            if args.ages:
                age_bin_list = [0, 15, 25, 35, 45, 55, 100]

            plot_population_by_age(dir_or_filename=dir_or_filename,
                                   exp_dir_or_filename=exp_dir_or_filename,
                                   node_id=node_id,
                                   gender=None,
                                   age_bin_list=age_bin_list,
                                   show_avg_per_run=args.mean,
                                   img_dir=args.output)

        elif plot == "prevalence":
            if args.expected:
                print("===============================================")
                print("Prevalence plot does not support expected data.")
                print("===============================================")

            age_bin_list = None
            if args.ages:
                age_bin_list = [15, 25, 35, 45, 55]

            plot_prevalence_for_dir(dir_or_filename=dir_or_filename,
                                    exp_dir_or_filename=exp_dir_or_filename,
                                    node_id=args.node_id,
                                    gender=None,
                                    age_bin_list=age_bin_list,
                                    show_avg_per_run=args.mean,
                                    show_fraction=False,
                                    img_dir=args.output)

        elif plot == "risk":
            if args.node_id is not None:
                print("===============================================")
                print("Population plot does not support node_id.")
                print("===============================================")

            age_bin_list = None
            if args.ages:
                age_bin_list = [15, 35, 55]

            plot_population_by_ip(dir_or_filename=dir_or_filename,
                                  exp_dir_or_filename=exp_dir_or_filename,
                                  node_id=args.node_id,
                                  ip_key="Risk",
                                  age_bin_list=age_bin_list,
                                  show_avg_per_run=args.mean,
                                  show_fraction=False,
                                  img_dir=args.output)
            # plot_risk_zambia(dir_or_filename=dir_or_filename,
            #                  age_bin_list=age_bin_list,
            #                  show_avg_per_run=args.mean,
            #                  show_fraction=True,
            #                  show_expected=args.expected,
            #                  img_dir=args.output)

        elif plot == "vmmc":
            age_bin_list = None
            if args.ages:
                age_bin_list = [15, 20, 25, 30, 35, 40, 45, 50, 55]

            plot_vmmc_by_age(dir_or_filename=dir_or_filename,
                             exp_dir_or_filename=exp_dir_or_filename,
                             node_id=args.node_id,
                             age_bin_list=age_bin_list,
                             show_avg_per_run=args.mean,
                             img_dir=args.output)
            # plot_vmmc_for_dir(dir_or_filename=dir_or_filename,
            #                   node_id=args.node_id,
            #                   age_bin_list=age_bin_list,
            #                   show_expected=args.expected,
            #                   show_avg_per_run=args.mean,
            #                   img_dir=args.output)

        elif plot == "art":
            if args.expected:
                print("========================================")
                print("ART plot does not support expected data.")
                print("========================================")

            age_bin_list = None
            if args.ages:
                age_bin_list = [0, 15, 50, 75, 100]

            plot_onART_by_age(dir_or_filename=dir_or_filename,
                              exp_dir_or_filename=exp_dir_or_filename,
                              node_id=args.node_id,
                              gender=None,
                              age_bin_list=age_bin_list,
                              show_avg_per_run=args.mean,
                              show_fraction=True,
                              fraction_of_infected=True,
                              img_dir=args.output)

        elif plot == "state":
            if args.expected:
                print("==========================================")
                print("State plot does not support expected data.")
                print("==========================================")

            age_bin_list = None
            if args.ages:
                age_bin_list = [15, 30, 45]

            plot_population_by_ip(dir_or_filename=dir_or_filename,
                                  exp_dir_or_filename=exp_dir_or_filename,
                                  node_id=args.node_id,
                                  ip_key="CascadeState",
                                  age_bin_list=age_bin_list,
                                  show_avg_per_run=args.mean,
                                  show_fraction=False,
                                  img_dir=args.output)

        elif plot == "column":
            plot_columns(filename=dir_or_filename,
                         title=args.column_name,
                         y_axis_name="Count",
                         column_names=[" " + args.column_name],
                         img_dir=args.output)

        else:
            raise ValueError(f"Plot type '{args.plot}' is not supported.")



from typing import Dict

from emod_api.demographics.updateable import Updateable


class ConcurrencyParameters(Updateable):
    def __init__(self,
                 probability_extra_relationship_male: float = 0,
                 probability_extra_relationship_female: float = 0,
                 max_simultaneous_relationships_male: float = 0,
                 max_simultaneous_relationships_female: float = 0):
        """
        A ConcurrencyParameters object represents the likelihood and maximum count of additional concurrent
        relationships for males and females in a simulation. One ConcurrencyParameters object typically is created
        per relationship type and risk group (e.g. one for Commercial HIGH risk relationships, one for
        Transitory MED risk relationships, etc.).

        https://docs.idmod.org/projects/emod-hiv/en/latest/parameter-demographics.html#concurrency-parameters

        Args:
            probability_extra_relationship_male: the probability of having extra relationships for males
            probability_extra_relationship_female: the probability of extra relationships for females
            max_simultaneous_relationships_male: the maximum simultaneous relationships for males
            max_simultaneous_relationships_female: the maximum simultaneous relationships for females
        """
        super().__init__()
        self.probability_extra_relationship_male = probability_extra_relationship_male
        self.probability_extra_relationship_female = probability_extra_relationship_female
        self.max_simultaneous_relationships_male = max_simultaneous_relationships_male
        self.max_simultaneous_relationships_female = max_simultaneous_relationships_female

    def to_dict(self) -> Dict:
        concurrency = {'Prob_Extra_Relationship_Male': self.probability_extra_relationship_male,
                       'Prob_Extra_Relationship_Female': self.probability_extra_relationship_female,
                       'Max_Simultaneous_Relationships_Male': self.max_simultaneous_relationships_male,
                       'Max_Simultaneous_Relationships_Female': self.max_simultaneous_relationships_female}
        return concurrency

    @classmethod
    def from_dict(cls, d: Dict) -> '__class__':
        return cls(probability_extra_relationship_male=d['Prob_Extra_Relationship_Male'],
                   probability_extra_relationship_female=d['Prob_Extra_Relationship_Female'],
                   max_simultaneous_relationships_male=d['Max_Simultaneous_Relationships_Male'],
                   max_simultaneous_relationships_female=d['Max_Simultaneous_Relationships_Female'])



from emod_api import campaign as api_campaign
from emodpy.campaign.base_intervention import IndividualIntervention
from emodpy.campaign.event import create_campaign_event

from emodpy_hiv.campaign.common import TargetDemographicsConfig, PropertyRestrictions, NChooserTargetedDistributionHIV, ValueMap
from emodpy_hiv.campaign.event_coordinator import NChooserEventCoordinatorHIV, ReferenceTrackingEventCoordinatorTrackingConfig
from emodpy_hiv.utils.emod_enum import TargetDiseaseState
from emodpy_hiv.utils.targeting_config import AbstractTargetingConfig

import pandas as pd
import warnings

# ported from emodpy/campaign/distributor.py

# This function adds the intervention(s) to the campaign at scheduled time with the given parameters.
# The start_year parameter is only allowed to be used in HIV, and it is recommended that HIV modelers use it to schedule
# the intervention(s) at the specified year instead of start_day.
from emodpy.campaign.distributor import add_intervention_scheduled

# This function configures the campaign to distribute an intervention to an individual when that individual broadcasts
# an event.
# The start_year parameter is only allowed to be used in HIV, and it is recommended that HIV modelers use it to define
# the specified year when the simulation starts to listen to the event instead of start_day.
from emodpy.campaign.distributor import add_intervention_triggered


def add_intervention_nchooser_df(campaign: api_campaign,
                                 intervention_list: list[IndividualIntervention],
                                 distribution_df: pd.DataFrame,
                                 property_restrictions: PropertyRestrictions = None,
                                 target_disease_state: list[list[TargetDiseaseState]] = None,
                                 target_disease_state_has_intervention_name: str = None,
                                 event_name: str = None,
                                 node_ids: list[int] = None) -> None:
    """
    # Overview:

    Distributes a list of individual-level interventions to exactly N people of a targeted demographic in HIV
    simulations. This contrasts with other event coordinators that distribute an intervention to a percentage of the
    population, not to an exact count.

    # Population Scaling:

    'N' is assumed to be a number that is for the non-scaled population. This means you can make it the number actually
    used in the real world. The 'N' values entered will be multiplied by the config parameter `x_Base_Population`.

    For example:
        - If you wanted to model how the real world distributed 2,000 male circumcisions in 1990 when the total
          population was 400,000, and your demographic parameters are configured such that EMOD has about 400,000 people
          in 1990, you would distribute 2,000 male circumcisions.
        - If the simulation is taking a long time, and you change `config.x_Base_Population` to 0.5 to cut the agents in
          the simulation in half, you would also want to reduce the number of male circumcisions being distributed.
          This feature will automatically adjust the number being targeted by `config.x_Base_Population` to only hand
          out 1,000 male circumcisions.

    # Distribution Over Time:

    NChooser will also spread out the number of interventions being distributed over the entire time period.

    For example:
        - Let's assume that you are trying to distribute MaleCircumcision to 9 men over five update periods. NChooser
          will give out two interventions the first four update periods and one during the last update period.

    # DataFrame Requirements:

    This function takes in a DataFrame containing the distribution data:
        - Required columns: `year`, `min_age`, `max_age`
        - At least one of the following columns is required: `num_targeted`, `num_targeted_female`, `num_targeted_male`

    The data in the DataFrame is used to create a list of `NChooserTargetedDistributionHIV` objects specifying:
        - When: with the `year` data
        - To whom: with the `min_age` and `max_age` data
        - How many interventions are distributed: with the `num_targeted` data, or the `num_targeted_female` for female
          individuals and `num_targeted_male` for male individuals

    Args:
        campaign (emod_api.campaign, required):
            - The campaign object to which the event will be added.
            - This should be an instance of the emod_api.campaign class.
        intervention_list (list[IndividualIntervention], required):
            - A list of IndividualIntervention objects. NodeIntervention is not supported in this feature.
            - Refer to the emodpy_hiv.campaign.individual_intervention module for available IndividualIntervention derived classes.
        distribution_df (pd.DataFrame, required):
            - A DataFrame containing the data for these columns: year, min_age, max_age, num_targeted,
              num_targeted_female, num_targeted_male.
            - The first three columns are required, and at least one of the last three columns is required.
            - num_targeted: The number of individuals to be targeted, gender-agnostic. It can't be used with
              num_targeted_female or num_targeted_male.
            - num_targeted_female: The number of female individuals to be targeted. It can't be used with num_targeted.
            - num_targeted_male: The number of male individuals to be targeted. It can't be used with num_targeted.
        target_disease_state (list[list[TargetDiseaseState]], optional):
            - A two-dimensional list of disease states using the TargetDiseaseState enum.
            - To qualify for the intervention, an individual must match at least one of the inner lists of disease states.
            - Each entry in the inner array is an "and" requirement, meaning an individual must match all states in one
              inner list to qualify.
            - Each inner list is combined with an "or" relationship with other inner list, meaning an individual needs
              to match at least one of the inner list to qualify.
            - Possible values: Refer to the TargetDiseaseState enum for the possible values.
            - Default value: None.
        target_disease_state_has_intervention_name (str, optional):
            - The name of the intervention to look for in an individual when targeting specific disease states.
            - It's required when using TargetDiseaseState.HAS_INTERVENTION or TargetDiseaseState.NOT_HAS_INTERVENTION in target_disease_state.
        property_restrictions (PropertyRestrictions, optional):
            - A PropertyRestrictions to define the individual-level property restrictions in the coordinator.
            - Please note that node property restrictions are not supported in this feature.
            - Default value: None.
        event_name (str, optional):
            - The name of the campaign event.
            - Default value: None
        node_ids (list[int], optional):
            - A list of node IDs where the event will be applied.
            - If None, the event applies to all nodes.
            - Default value: None.

    Returns:
            None: This function does not return anything. It modifies the campaign object in place.

    Examples:
        Example 1: This example demonstrates how to distribute a MaleCircumcision intervention to male individuals
        who are HIV positive and do not have the intervention already. The intervention is distributed based on the
        distribution data provided in a DataFrame.

        >>> import emod_api
        >>> from emodpy_hiv.campaign.distributor import add_intervention_nchooser_df
        >>> from emodpy_hiv.campaign.individual_intervention import MaleCircumcision
        >>> from emodpy_hiv.campaign.common import CommonInterventionParameters as CIP
        >>> from emodpy_hiv.utils.emod_enum import TargetDiseaseState
        >>> import pandas as pd
        >>>
        >>> campaign_obj = emod_api.campaign
        >>> campaign_obj.schema_path = 'path_to_schema'
        >>> # Initialize a MaleCircumcision intervention with intervention_name: 'MaleCircumcision'
        >>> intervention_name='MaleCircumcision'
        >>> mc = MaleCircumcision(campaign_obj, common_intervention_parameters=CIP(intervention_name=intervention_name)
        >>> # Create a DataFrame with the distribution data: year, min_age, max_age, num_targeted_male
        >>> # The intervention will be distributed to MALE individuals with the following values:
        >>> #    for the year 2010, age ranges: [1, 14.999), [15, 49.999), the number of targeted MALE individuals are: [200, 1300].
        >>> #    for the year 2011, age ranges: [1, 14.999), [15, 49.999), the number of targeted MALE individuals are: [290, 1490].
        >>> data = {'year': [2010, 2010, 2011, 2011],
        >>>         'min_age': [1, 15, 1, 15],
        >>>         'max_age': [14.999, 49.999, 14.999, 49.999],
        >>>         'num_targeted_male': [200, 1300, 290, 1490]}
        >>> distributions_df = pd.DataFrame.from_dict(data)
        >>> # Distribute the MaleCircumcision intervention to the campaign with the distribution data. The targeted
        >>> # individuals should be male, HIV negative and don't have an intervention called 'MaleCircumcision' already.
        >>> add_intervention_nchooser_df(campaign_obj,
        >>>                              intervention_list=[mc],
        >>>                              target_disease_state=[[TargetDiseaseState.HIV_POSITIVE, TargetDiseaseState.NOT_HAVE_INTERVENTION]],
        >>>                              target_disease_state_has_intervention_name=intervention_name,
        >>>                              distribution_df=distributions_df)

        Example 2: This example demonstrates the usage of "And" and "Or" relationship in the target_disease_state parameter.
        With the following target_disease_state parameter, the MaleCircumcision intervention will be distributed to
        individuals who don't have the intervention already and are either HIV positive or tested positive.

        >>> add_intervention_nchooser_df(campaign_obj,
        >>>                              intervention_list=[mc],
        >>>                              target_disease_state=[[TargetDiseaseState.HIV_POSITIVE, TargetDiseaseState.NOT_HAVE_INTERVENTION],
        >>>                                                   [TargetDiseaseState.TESTED_POSITIVE, TargetDiseaseState.NOT_HAVE_INTERVENTION]],
        >>>                              target_disease_state_has_intervention_name=intervention_name,
        >>>                              distribution_df=distributions_df)

    Raises:

    """
    # These are the columns that are expected in the distribution dataframe
    required_columns = {'year', 'min_age', 'max_age'}
    if not required_columns.issubset(distribution_df.columns):
        raise ValueError(f"Expected these columns: {required_columns} in the distribution_df "
                         f"dataframe. Got distributions.columns: {distribution_df.columns}")

    # Only one of the following names is required.
    optional_columns = {'num_targeted', 'num_targeted_female', 'num_targeted_male'}
    # Check if at least one optional column is present in the distribution dataframe
    if not optional_columns.intersection(distribution_df.columns):
        raise ValueError(f"Expected at least one of these columns: {optional_columns} in the distribution_df "
                         f"dataframe. Got distributions.columns: {distribution_df.columns}")

    # Get the unique value of target years from the distribution dataframe
    target_years = distribution_df.year.unique()
    # Created the distribution objects for each target year
    targeted_distributions = []
    for year in target_years:
        # Filter the distribution dataframe with year
        df_cur_year = distribution_df[distribution_df['year'] == year].sort_values(by=['min_age'])
        # Max and Min age ranges for qualifying individuals.
        age_ranges_years = [df_cur_year['min_age'].tolist(), df_cur_year['max_age'].tolist()]
        num_targeted = None
        num_targeted_female = None
        num_targeted_male = None
        if 'num_targeted' in distribution_df.columns:
            num_targeted = df_cur_year['num_targeted'].tolist()
            if 'num_targeted_female' in distribution_df.columns or 'num_targeted_male' in distribution_df.columns:
                raise ValueError("num_targeted column should not be used with num_targeted_female or num_targeted_male.")
        else:
            if 'num_targeted_female' in distribution_df.columns:
                num_targeted_female = df_cur_year['num_targeted_female'].tolist()
            if 'num_targeted_male' in distribution_df.columns:
                num_targeted_male = df_cur_year['num_targeted_male'].tolist()

            # If one of them is not set, it should contain 0s with the same length as the other one.
            if num_targeted_female is None:
                num_targeted_female = [0] * len(num_targeted_male)
            if num_targeted_male is None:
                num_targeted_male = [0] * len(num_targeted_female)

        targeted_distribution = NChooserTargetedDistributionHIV(age_ranges_years=age_ranges_years,
                                                                start_year=float(year),
                                                                end_year=year + 0.999999,
                                                                num_targeted=num_targeted,
                                                                num_targeted_females=num_targeted_female,
                                                                num_targeted_males=num_targeted_male,
                                                                property_restrictions=property_restrictions,
                                                                target_disease_state=target_disease_state,
                                                                target_disease_state_has_intervention_name=target_disease_state_has_intervention_name)
        targeted_distributions.append(targeted_distribution)

    # Call the add_intervention_nchooser function to distribute the intervention_list with target_distributions
    # Set start_year to the minimum year in the target_years
    start_year = float(min(target_years))
    _add_intervention_nchooser(campaign,
                               intervention_list=intervention_list,
                               targeted_distributions=targeted_distributions,
                               start_year=start_year,
                               event_name=event_name,
                               node_ids=node_ids)


def _add_intervention_nchooser(campaign: api_campaign,
                               intervention_list: list[IndividualIntervention],
                               targeted_distributions: list[NChooserTargetedDistributionHIV],
                               start_year: float,
                               event_name: str = None,
                               node_ids: list[int] = None) -> None:
    """
    Distributes a list of individual-level interventions to exactly N people of a targeted demographic in HIV
    simulations. This contrasts with other event coordinators that distribute an intervention to a percentage of the
    population, not to an exact count.

    This function takes in a list of NChooserTargetedDistributionHIV object specifying when, to whom, and how many
    interventions are distributed.

    Please refer to the Emod documentation for NChooserEventCoordinatorHIV at the following link:
    :doc:`emod-hiv:parameter-campaign-event-nchoosereventcoordinatorhiv`

    Args:
        campaign (api_campaign, required):
            - The campaign object to which the event will be added.
            - This should be an instance of the emod_api.campaign class.
        intervention_list (list[IndividualIntervention], required):
            - A list of IndividualIntervention objects. NodeIntervention is not supported in this feature.
            - Refer to the emodpy_hiv.campaign.individual_intervention module for available IndividualIntervention derived classes.
        targeted_distributions (list[NChooserTargetedDistributionHIV], required):
            - A list of NChooserTargetedDistributionHIV object specifying when, to whom, and how many interventions are distributed.
            - Please refer emodpy_hiv.campaign.event_coordinator.NChooserTargetedDistributionHIV for more details.
        start_year (float, requied):
            - The year when the event starts.
        event_name (str, optional):
            - The name of the campaign event.
            - Default value: None.
        node_ids (list[int], optional):
            - A list of node IDs where the event will be applied.
            - If None, the event applies to all nodes.
            - Default value: None.

    Returns:
        None: This function does not return anything. It modifies the campaign object in place.

    Examples:
        >>> from emodpy_hiv.campaign.common import NChooserTargetedDistributionHIV
        >>> from emod_api import campaign as api_campaign
        >>> from emodpy_hiv.campaign.individual_intervention import MaleCircumcision
        >>> from emodpy_hiv.campaign.distributor import _add_intervention_nchooser
        >>>
        >>> api_campaign.set_schema('path_to_schema')
        >>> intervention_config = MaleCircumcision(api_campaign)
        >>> distribution_1 = NChooserTargetedDistributionHIV( ...)
        >>> distribution_2 = NChooserTargetedDistributionHIV( ...)
        >>> _add_intervention_nchooser(api_campaign,
        >>>                           intervention_list=[intervention_config],
        >>>                           targeted_distributions=[distribution_1, distribution_2],
        >>>                           start_year=1990,
        >>>                           node_ids=[1, 2, 3])
    """
    # Create a NChooserEventCoordinatorHIV with the intervention
    coordinator = NChooserEventCoordinatorHIV(campaign, intervention_list=intervention_list,
                                              targeted_distributions=targeted_distributions)
    # Create a CampaignEventByYear object with the coordinator
    event = create_campaign_event(campaign, coordinator=coordinator, event_name=event_name, node_ids=node_ids,
                                  start_day=None, start_year=start_year)

    # Add the event to the campaign
    campaign.add(event.to_schema_dict(campaign))


def add_intervention_reference_tracking(campaign: api_campaign,
                                        intervention_list: list[IndividualIntervention],
                                        time_value_map: ValueMap,
                                        tracking_config: AbstractTargetingConfig,
                                        start_year: float,
                                        end_year: float = 2200,
                                        update_period: float = 365,
                                        target_demographics_config: TargetDemographicsConfig = TargetDemographicsConfig(demographic_coverage=None),
                                        property_restrictions: PropertyRestrictions = None,
                                        targeting_config: AbstractTargetingConfig = None,
                                        event_name: str = None,
                                        node_ids: list[int] = None) -> None:
    """
    Distribute interventions to the population such that a user determined coverage of an attribute is maintained over
    time.

    This function creates a "tracker" that will track the prevalence of a specific attribute in the population and
    distribute interventions to achieve the desired coverage. The tracker will periodically poll the population to
    determine the current prevalence of the attribute and distribute the number of interventions needed to get
    prevalence up to the desired value. If prevalence is higher than the desired value, then no interventions will be
    distributed. Other things outside of this tracker (like expiration timers in the intervention) maybe needed to
    cause the coverage to stay down to the desired coverage.

    Args:
        campaign (api_campaign, required):
            - The campaign object to which the intervention will be added.
        intervention_list (list[IndividualIntervention], required):
            - A list of IndividualIntervention objects. NodeIntervention is not supported in this feature.
            - Refer to the emodpy_hiv.campaign.individual_intervention module for available IndividualIntervention derived classes.
        time_value_map (ValueMap, required):
            - A ValueMap object that defines the target coverage levels over time.
        tracking_config (AbstractTargetingConfig, required):
            - An AbstractTargetingConfig object that defines the attribute to be tracked within the targeted group.
            - The number of individuals with this attribute is used as the numerator for coverage calculations.
            - The number of individual selected via targeting_config, target_demographics, and property_restrictions determine the denominator.
            - Please refer to the emodpy_hiv.utils.targeting_config module for more information.
        start_year (float, required):
            - The year to start distributing the intervention.
            - Defines the start of the time period over which the intervention will be distributed.
            - Minimum value: 1900,
            - Maximum value: 2200.
        end_year (float, optional):
            - The year to stop distributing the intervention.
            - Defines the end of the time period over which the intervention will be distributed.
            - Minimum value: 1900,
            - Maximum value: 2200,
            - Default value: 2200.
        update_period (float, optional):
            - The time period in days between when the tracker polls the population and determines if it should
              distribute more interventions.
            - Minimum value: 1,
            - Maximum value: 3650,
            - Default value: 365.
        target_demographics_config (TargetDemographicsConfig, optional):
            - A TargetDemographicsConfig to define the demographic characteristics of individuals targeted by the intervention.
            - Please refer to the emodpy_hiv.campaign.common.TargetDemographicsConfig module for more information.
            - Please note that the demographic_coverage is not used in this coordinator. Use TargetDemographicsConfig(demographic_coverage=None) to avoid errors.
            - Default value: TargetDemographicsConfig(demographic_coverage=None).
        property_restrictions (PropertyRestrictions, optional):
            - A PropertyRestrictions object that defines individual-level property restrictions for the intervention.
            - Node property restrictions are not supported.
            - Default value: None.
        targeting_config (AbstractTargetingConfig, optional):
            - An AbstractTargetingConfig to define who to target individuals with the intervention besides the
              target_demographics_config and property_restrictions.
            - Please note that when defining the target group, you do NOT want to specify the negative what you want to
              track. because the number of people in the target group is your denominator when calculating coverage.
              For example, if you want the tracker to have 50% of men circumcised, then you need men who are both
              circumcised and not circumcised in the denominator and only the number circumcised men in the numerator.
            - Please refer to the emodpy_hiv.utils.targeting_config module for more information.
            - Default value: None.
        event_name (str, optional):
            - The name of the campaign event.
            - Default value: None.
        node_ids (list[int], optional):
            - A list of node IDs to which the intervention should be applied.
            - If None, the intervention is applied to all nodes.
            - Default value: None.

    Returns:
        None: This function does not return anything. It modifies the campaign object in place.

    Examples:
        Use a reference tracker to ensure that the fraction of medium risk men that are circumcised meets certain
        levels each year from 1960 through 1965 where the tracker is updating who is circumcised every 6 months (182 days)
        If coverage is below the target level at the time of polling, apply the MaleCircumcision
        intervention to uncircumcised men to reach the target coverage.

        Please note that you don't need to specify the negative of what you want to track(~IsCircumcised()) in the
        targeting_config. See more details in the targeting_config argument description.

        >>> import emod_api
        >>> from emodpy_hiv.campaign.distributor import add_intervention_reference_tracking
        >>> from emodpy_hiv.campaign.individual_intervention import MaleCircumcision
        >>> from emodpy_hiv.campaign.common import (ValueMap, TargetGender, CommonInterventionParameters as CIP,
        >>>                                        TargetDemographicsConfig as TDC)
        >>> from emodpy_hiv.utils.targeting_config import IsCircumcised, HasIP
        >>>
        >>> campaign_obj = emod_api.campaign
        >>> campaign_obj.schema_path = 'path_to_schema'
        >>> mc = MaleCircumcision(campaign_obj,
        >>>                       distributed_event_trigger='VMMC_1')
        >>> time_value_map = ValueMap(times=[1960,  1961,   1962,   1963,   1964],
        >>>                          values=[0.25,  0.375,  0.4,    0.4375, 0.46875])
        >>> targeting_config = HasIP(ip_key_value="Risk:MEDIUM")
        >>> tracking_config = IsCircumcised()
        >>> add_intervention_reference_tracking(campaign_obj,
        >>>                                     intervention_list=[mc],
        >>>                                     time_value_map=time_value_map,
        >>>                                     tracking_config=tracking_config,
        >>>                                     targeting_config=targeting_config,
        >>>                                     start_year=1960,
        >>>                                     end_year=1965,
        >>>                                     update_period=182,
        >>>                                     target_demographics_config=TDC(target_gender=TargetGender.MALE))
    """
    if not isinstance(time_value_map, ValueMap):
        raise ValueError("The time_value_map should be an instance of the ValueMap class.")
    # Validate the start_year and end_year
    if end_year <= start_year:
        raise ValueError("The end_year should be greater than the start_year.")
    # All years in ValueMap should be within the start_year and end_year
    if all([year < start_year or year > end_year for year in time_value_map._times]):
        raise ValueError("All years in the time_value_map is not within the start_year and end_year.")
    # At lease one year in the ValueMap is not within the start_year and end_year
    if any([year < start_year or year > end_year for year in time_value_map._times]):
        warnings.warn("At least one year in the time_value_map is not be within the start_year and end_year."
                      "Please make sure that is intended.\n"
                      f"start_year: {start_year}, end_year: {end_year}, years in time_value_map: {time_value_map._times}.",
                      category=UserWarning)

    # Create a ReferenceTrackingEventCoordinatorTrackingConfig with the intervention_list
    coordinator = ReferenceTrackingEventCoordinatorTrackingConfig(campaign,
                                                                  intervention_list=intervention_list,
                                                                  time_value_map=time_value_map,
                                                                  targeting_config=targeting_config,
                                                                  tracking_config=tracking_config,
                                                                  end_year=end_year,
                                                                  update_period=update_period,
                                                                  target_demographics_config=target_demographics_config,
                                                                  property_restrictions=property_restrictions,
                                                                  )
    # Create a CampaignEventByYear object with the coordinator
    event = create_campaign_event(campaign, coordinator=coordinator, event_name=event_name, node_ids=node_ids,
                                  start_day=None, start_year=start_year)

    # Add the event to the campaign
    campaign.add(event.to_schema_dict(campaign))


# __all_exports: A list of classes that are intended to be exported from this module.
__all_exports = [add_intervention_scheduled, add_intervention_triggered, _add_intervention_nchooser,
                 add_intervention_nchooser_df, add_intervention_reference_tracking]

# The following loop sets the __module__ attribute of each class in __all_exports to the name of the current module.
# This is done to ensure that when these classes are imported from this module, their __module__ attribute correctly
# reflects their source module.
# During the documentation build with Sphinx, these classes will be displayed as belonging to the 'emodpy_hiv' package,
# not the 'emodpy' package.
# For example, the 'PropertyRestrictions' class will be documented as 'emodpy_hiv.campaign.common.PropertyRestrictions(...)'.
# This is crucial for accurately representing the source of these classes in the documentation.

for _ in __all_exports:
    _.__module__ = __name__

# __all__: A list that defines the public interface of this module.
# This is essential to ensure that Sphinx builds documentation for these classes, including those that are imported
# from emodpy.
# It contains the names of all the classes that should be accessible when this module is imported using the syntax
# 'from module import *'.
# Here, it is set to the names of all classes in __all_exports.

__all__ = [_.__name__ for _ in __all_exports]



from typing import Dict, Set

from emodpy_hiv.demographics.assortivity import Assortivity
from emodpy_hiv.demographics.concurrency_parameters import ConcurrencyParameters
from emodpy_hiv.demographics.pair_formation_parameters import PairFormationParameters
from emodpy_hiv.demographics.relationship_parameters import RelationshipParameters


class Society:
    def __init__(self,
                 concurrency_configuration: Dict = None,
                 relationship_parameters: Dict[str, RelationshipParameters] = None,
                 pair_formation_parameters: Dict[str, PairFormationParameters] = None,
                 concurrency_parameters: Dict[str, Dict[str, ConcurrencyParameters]] = None):
        """
        The Society object defines the behavioral-based parameters of a relationship type in the STI and HIV models,
        such as rates of partnership formation, partner preference, relationship duration, and concurrent partnerships.
        It must contain the (argument listed)) sets of relationship type parameters and the concurrency configuration.

        https://docs.idmod.org/projects/emod-hiv/en/latest/parameter-demographics.html#society

        Args:
            concurrency_configuration: a non-default concurrency configuration dict, if provided. A broadly applicable
                default is provided in method _default_concurrency_configuration() .
            relationship_parameters: a dict of RelationshipParameters objects, each key: value pair being
                relationship_type: RelationshipParameters()
            pair_formation_parameters: a dict of PairFormationParameters objects, each key: value pair being
                relationship_type: PairFormationParameters()
            concurrency_parameters: a nested dict of ConcurrencyParameters objects, indexed by type then risk, e.g.:
            {
                "INFORMAL": {
                    "HIGH": ConcurrencyParameter object,
                    "MED": ConcurrencyParameter object,
                    "LOW": ConcurrencyParameter object
                },
                "TRANSITORY": {
                    "HIGH": ConcurrencyParameter object,
                    "MED": ConcurrencyParameter object,
                    "LOW": ConcurrencyParameter object
                },
                ...
            }
        """
        super().__init__()
        self.concurrency_configuration = concurrency_configuration if concurrency_configuration is not None \
            else self._default_concurrency_configuration
        self.relationship_parameters = relationship_parameters if relationship_parameters is not None else {}
        self.pair_formation_parameters = pair_formation_parameters if pair_formation_parameters is not None else {}
        self.concurrency_parameters = concurrency_parameters if concurrency_parameters is not None else {}

    @property
    def relationship_types(self) -> Set[str]:
        return {*self.relationship_parameters.keys(),
                *self.pair_formation_parameters.keys(),
                *self.concurrency_parameters.keys()}

    @property
    def _default_concurrency_configuration(self):
        return {
            "Probability_Person_Is_Behavioral_Super_Spreader": 0,
            "Individual_Property_Name": "Risk",
            "LOW": {
                "Extra_Relational_Flag_Type": "Correlated",
                "Correlated_Relationship_Type_Order": [
                    "COMMERCIAL",
                    "TRANSITORY",
                    "INFORMAL",
                    "MARITAL"
                ]
            },
            "MEDIUM": {
                "Extra_Relational_Flag_Type": "Correlated",
                "Correlated_Relationship_Type_Order": [
                    "COMMERCIAL",
                    "TRANSITORY",
                    "INFORMAL",
                    "MARITAL"
                ]
            },
            "HIGH": {
                "Extra_Relational_Flag_Type": "Correlated",
                "Correlated_Relationship_Type_Order": [
                    "COMMERCIAL",
                    "TRANSITORY",
                    "INFORMAL",
                    "MARITAL"
                ]
            }
        }

    def get_pair_formation_parameters_by_relationship_type(self, relationship_type: str) -> PairFormationParameters:
        if relationship_type not in self.pair_formation_parameters:
            self.pair_formation_parameters[relationship_type] = PairFormationParameters()
        return self.pair_formation_parameters[relationship_type]

    def get_concurrency_parameters_by_relationship_type_and_risk(self, relationship_type: str,
                                                                 risk: str) -> ConcurrencyParameters:
        if relationship_type not in self.concurrency_parameters:
            self.concurrency_parameters[relationship_type] = {}
        if risk not in self.concurrency_parameters[relationship_type]:
            self.concurrency_parameters[relationship_type][risk] = ConcurrencyParameters()
        return self.concurrency_parameters[relationship_type][risk]

    def get_relationship_parameters_by_relationship_type(self, relationship_type: str) -> RelationshipParameters:
        if relationship_type not in self.relationship_parameters:
            self.relationship_parameters[relationship_type] = RelationshipParameters()
        return self.relationship_parameters[relationship_type]

    def set_pair_formation_parameters(self, relationship_type: str,
                                      formation_rate: float = None,
                                      assortivity: Assortivity = None) -> None:
        pair_formation_parameters = self.get_pair_formation_parameters_by_relationship_type(relationship_type=relationship_type)
        if formation_rate is not None:
            pair_formation_parameters.Formation_Rate_Constant = formation_rate
        if assortivity is not None:
            if len(assortivity.matrix) != 3:
                raise ValueError("You need to have 3 rows in your assortivity matrix.")
            if any(len(row) != 3 for row in assortivity.matrix):
                raise ValueError("You need to have 3 columns in each row of your assortivity matrix.")
            pair_formation_parameters.Assortivity = assortivity

    def set_concurrency_parameters(self, relationship_type: str,
                                   risk: str,
                                   max_simul_rels_male: float = None,
                                   max_simul_rels_female: float = None,
                                   prob_xtra_rel_male: float = None,
                                   prob_xtra_rel_female: float = None) -> None:
        concurrency_parameters = self.get_concurrency_parameters_by_relationship_type_and_risk(relationship_type=relationship_type, risk=risk)
        if max_simul_rels_male is not None:
            concurrency_parameters.max_simultaneous_relationships_male = max_simul_rels_male
        if max_simul_rels_female is not None:
            concurrency_parameters.max_simultaneous_relationships_female = max_simul_rels_female
        if prob_xtra_rel_male is not None:
            concurrency_parameters.probability_extra_relationship_male = prob_xtra_rel_male
        if prob_xtra_rel_female is not None:
            concurrency_parameters.probability_extra_relationship_female = prob_xtra_rel_female

    def set_relationship_parameters(self, relationship_type: str,
                                    coital_act_rate: float = None,
                                    condom_usage_min: float = None,
                                    condom_usage_mid: float = None,
                                    condom_usage_max: float = None,
                                    condom_usage_rate: float = None,
                                    duration_scale: float = None,
                                    duration_heterogeneity: float = None) -> None:
        relationship_parameters = self.get_relationship_parameters_by_relationship_type(relationship_type=relationship_type)
        if coital_act_rate is not None:
            relationship_parameters.coital_act_rate = coital_act_rate
        if condom_usage_min is not None:
            relationship_parameters.condom_usage.min = condom_usage_min
        if condom_usage_mid is not None:
            relationship_parameters.condom_usage.mid = condom_usage_mid
        if condom_usage_max is not None:
            relationship_parameters.condom_usage.max = condom_usage_max
        if condom_usage_rate is not None:
            relationship_parameters.condom_usage.rate = condom_usage_rate
        if duration_scale is not None:
            relationship_parameters.duration.weibull_lambda = duration_scale
        if duration_heterogeneity is not None:
            relationship_parameters.duration.weibull_kappa = duration_heterogeneity ** -1

    def to_dict(self) -> Dict:
        society = {'Concurrency_Configuration': self.concurrency_configuration}
        for relationship_type in self.relationship_types:
            relationship_configuration = {}

            relationship_parameters = self.relationship_parameters.get(relationship_type, None)
            if relationship_parameters is not None:
                relationship_configuration['Relationship_Parameters'] = relationship_parameters.to_dict()

            pair_formation_parameters = self.pair_formation_parameters.get(relationship_type, None)
            if pair_formation_parameters is not None:
                relationship_configuration['Pair_Formation_Parameters'] = pair_formation_parameters.to_dict()

            concurrency_parameters = self.concurrency_parameters.get(relationship_type, None)
            if concurrency_parameters is not None:
                if 'Concurrency_Parameters' not in relationship_configuration:
                    relationship_configuration['Concurrency_Parameters'] = {}
                for risk_group, cp_by_risk in concurrency_parameters.items():
                    relationship_configuration['Concurrency_Parameters'][risk_group] = cp_by_risk.to_dict()

            society[relationship_type] = relationship_configuration
        return society

    @classmethod
    def from_dict(cls, d: Dict) -> '__class__':
        concurrency_configuration = d['Concurrency_Configuration']
        relationship_parameters = {}
        pair_formation_parameters = {}
        concurrency_parameters = {}

        relationship_types = d.keys() - ['Concurrency_Configuration']
        for relationship_type in relationship_types:
            relationship_params = d[relationship_type].get('Relationship_Parameters', None)
            if relationship_params is not None:
                relationship_parameters[relationship_type] = RelationshipParameters.from_dict(relationship_params)

            pair_formation_params = d[relationship_type].get('Pair_Formation_Parameters', None)
            if pair_formation_params is not None:
                pair_formation_parameters[relationship_type] = PairFormationParameters.from_dict(pair_formation_params)

            concurrency_params = d[relationship_type].get('Concurrency_Parameters', None)
            if concurrency_params is not None:
                cp_by_risk = {risk: ConcurrencyParameters.from_dict(by_risk_parameters)
                              for risk, by_risk_parameters in concurrency_params.items()}
                concurrency_parameters[relationship_type] = cp_by_risk

        return cls(concurrency_configuration=concurrency_configuration, relationship_parameters=relationship_parameters,
                   pair_formation_parameters=pair_formation_parameters, concurrency_parameters=concurrency_parameters)



from emodpy.campaign.waning_config import AbstractWaningConfig, BaseWaningConfig, Constant, MapLinear, MapPiecewise, RandomBox, \
    MapLinearAge, MapLinearSeasonal, Box, BoxExponential, Exponential, Combo


# __all_exports: A list of classes that are intended to be exported from this module.
# These classes include the BaseWaningConfig, Constant, MapLinear, MapPiecewise, RandomBox, MapLinearAge,
# MapLinearSeasonal, Box, BoxExponential, Exponential, and Combo classes.

__all_exports = [
    AbstractWaningConfig,
    BaseWaningConfig,
    Constant,
    MapLinear,
    MapPiecewise,
    RandomBox,
    MapLinearAge,
    MapLinearSeasonal,
    Box,
    BoxExponential,
    Exponential,
    Combo]

# The following loop sets the __module__ attribute of each class in __all_exports to the name of the current module.
# This is done to ensure that when these classes are imported from this module, their __module__ attribute correctly
# reflects their source module.
# During the documentation build with Sphinx, these classes will be displayed as belonging to the 'emodpy_hiv' package,
# not the 'emodpy' package.
# For example, the 'MapLinear' class will be documented as 'emodpy_hiv.campaign.waning_config.MapLinear(...)'.
# This is crucial for accurately representing the source of these classes in the documentation.

for _ in __all_exports:
    _.__module__ = __name__

# __all__: A list that defines the public interface of this module.
# This is essential to ensure that Sphinx builds documentation for these classes, including those that are imported
# from emodpy.
# It contains the names of all the classes that should be accessible when this module is imported using the syntax
# 'from module import *'.
# Here, it is set to the names of all classes in __all_exports.

__all__ = [_.__name__ for _ in __all_exports]



def download(experiment_id, local_output_path="", files_to_get=None):
    """
    Download HIV output file(s) to local disk. Just gets ReportHIVByAgeAndGender.csv. Intended to be
    used command line: python -m emopyhiv.download <COMPS Experiment ID> <Optional Output Path> <Files To Get>
    Note that target files need to include "output/" if that's where they are.
    """
    from idmtools.core.platform_factory import Platform
    platform = Platform("Calculon", node_group="idm_48cores", priority="Highest")
    from idmtools_platform_comps.utils.download.download import DownloadWorkItem
    if files_to_get is None:
        files_to_get = ["output/ReportHIVByAgeAndGender.csv"]
    elif type(files_to_get) is str:
        files_to_get = files_to_get.split(',')

    print(f"Attempting to download {files_to_get} for experiment {experiment_id} into {local_output_path}.")
    dl_wi = DownloadWorkItem(
        related_experiments=[experiment_id],
        file_patterns=files_to_get,
        output_path=local_output_path,
    )
    dl_wi.run(wait_until_done=True, platform=platform)


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        print("Usage: python -m emopyhiv.download <COMPS Experiment ID> <Optional Output Path> <Optional List Of Files To Get>")
        sys.exit()
    download(sys.argv[1], (sys.argv[2] if len(sys.argv) > 2 else ""), (sys.argv[3] if len(sys.argv) == 4 else None))



from enum import Enum


class RiskGroups(Enum):
    """
    Aid to reduce direct string comparisons of standard HIV-model relationship types and associated errors.
    e.g. Use RiskGroups.high or RiskGroups.high.name in code depending on use and
    RiskGroups.high.value in json/dict values.
    """
    low = 'LOW'
    medium = 'MEDIUM'
    high = 'HIGH'



"""
This module contains stuff related to a YearAgeRate dataframe object that can be used
to create demographic objects.
"""

from typing import List, Tuple, Dict
import pandas as pd
from emod_api.demographics.age_distribution import AgeDistribution
from emod_api.demographics.fertility_distribution import FertilityDistribution
from emod_api.demographics.mortality_distribution import MortalityDistribution


class YearAgeRate:
    """
    The YearAgeRate class is a wrapper around a pandas dataframe such that the dataframe
    is expected to have a specific format that is used in the demographics.  Objects of
    this class are used as output from UN World Pop functions and as input into the
    creation of a Demograhics object.  This gives us a standard format of the data for
    most population-based demographic data.

    The dataframe is expected to have four columns:
        * node_id: An integer representing the node the data is associated with
        * min_year: A float representing the calendar start year of the period of years
            for which the represents.
        * min_age: A float representing the age of the people in years.
        * rate: The value (a float) of this column depends on the data contained.
            It can be a rate of fertility or mortality or a fraction of the population.

    This format assumes that the year/age ranges are contiguous from one value to the
    next largest value.  For example, if a row has a min_age = 15 and the next row of
    the same year with the next largest min_age is 20, then the data for the row where
    min_age = 15 is for the age range of [15-20), where 15 is included and 20 is excluded.
    This is the same for min_year.

    For a given node id, the format assumes that each min_year has the exact same set
    of min_ages.  Different nodes can have different min_years or min_ages, but within
    a given node, they must be the same.  The format also assumes that min_year and
    min_age for a given node are not duplicated.
    """

    COL_NAME_NODE_ID  = "node_id"    # Noqa: E221
    COL_NAME_MIN_YEAR = "min_year"
    COL_NAME_MIN_AGE  = "min_age"    # Noqa: E221
    COL_NAME_RATE     = "rate"       # Noqa: E221
    COL_NAMES = [
        COL_NAME_NODE_ID,
        COL_NAME_MIN_YEAR,
        COL_NAME_MIN_AGE,
        COL_NAME_RATE
    ]
    SORT_BY_COLUMNS = [
        COL_NAME_NODE_ID,
        COL_NAME_MIN_YEAR,
        COL_NAME_MIN_AGE
    ]

    @staticmethod
    def _validate_df(df):
        """
        Throws an exception if dataframe doesn't conform to one of the following:
            * Not the expected column names
            * Number of min_ages not the same for each min_year
            * Not the exact set of min_ages for each min_year
            * If there are dupliate rows that have the same node_id, min_year, and min_age.
        """
        node_group_collection = df.groupby(YearAgeRate.COL_NAME_NODE_ID)
        for node_id, node_group in node_group_collection:
            prev_min_year = 0
            prev_num_min_ages = 0
            prev_min_ages = []
            min_year_group_collection = node_group.groupby(YearAgeRate.COL_NAME_MIN_YEAR)
            for min_year, min_year_group in min_year_group_collection:
                # check that there are no duplicate min_years
                min_ages = sorted(min_year_group[YearAgeRate.COL_NAME_MIN_AGE].unique())
                num_min_ages_unique = len(min_ages)
                num_min_ages_total  = len(min_year_group[YearAgeRate.COL_NAME_MIN_AGE])              # Noqa: E221
                if num_min_ages_total != num_min_ages_unique:
                    min_year_list = min_year_group[YearAgeRate.COL_NAME_MIN_AGE].tolist()
                    msg =  f"Invalid duplicate number of entries for {YearAgeRate.COL_NAME_MIN_AGE} "# Noqa: E222
                    msg += f"for node_id={node_id} and min_year={min_year}.\n"
                    msg += f"min_ages={min_year_list}"
                    raise ValueError(msg)
                elif prev_num_min_ages == 0:
                    prev_num_min_ages = num_min_ages_total
                    prev_min_ages = min_ages
                    prev_min_year = min_year
                elif prev_num_min_ages != num_min_ages_total:
                    msg =  f"Invalid number of min_ages for min_year={min_year}.\n"                  # Noqa: E222
                    msg += f"Number of min_ages={prev_num_min_ages} for min_year={prev_min_year}.\n"
                    msg += f"Number of min_ages={num_min_ages_total} for min_year={min_year}."
                    raise ValueError(msg)
                elif min_ages != prev_min_ages:
                    msg =  f"Inconsistent set of min_ages for min_year={min_year}.\n"                # Noqa: E222
                    msg += f"Min_ages for min_year={prev_min_year} are: {prev_min_ages}.\n"
                    msg += f"Min_ages for min_year={min_year} are: {min_ages}."
                    raise ValueError(msg)

    def __init__(self,
                 df: pd.DataFrame = None,
                 csv_filename: str = None):
        """
        Construct a YearAgeRate object from the given dataframe OR one that is contained in a CSV.
        You can only define the dataframe OR the csv_filename, bot NOT both.

        Args:
            df: A pandas dataframe that follows the format of a YearAgeRate object.
            csv_filename: A CSV file containing data for YearAgeRate dataframe.
        """
        if (df is None and csv_filename is None) or (df is not None and csv_filename is not None):
            raise ValueError("You must define either 'df' or 'csv_filename', but not both.")

        if csv_filename is not None:
            df = pd.read_csv(csv_filename)

        if YearAgeRate.COL_NAMES in df.columns.tolist():
            msg = "The dataframe provided does not have the column names expected.\n"
            msg += f"Must have at least: {YearAgeRate.COL_NAMES}\n"
            msg += f"Actual: {df.columns.tolist()}"
            raise ValueError(msg)

        YearAgeRate._validate_df(df)

        self.df = df.sort_values(by=YearAgeRate.SORT_BY_COLUMNS, ascending=True)

    def to_csv(self, csv_filename):
        """
        Save the dataframe to a csv file.

        Args:
            csv_filename: The name of the file to write the dataframe to.
        """
        self.df.to_csv(csv_filename, index=False)

    def to_age_distributions(self) -> List[Tuple[int, AgeDistribution]]:
        """
        Convert this YearAgeRate object ot a list of (node_id, AgeDistribution) tuples.
        For each node in the dataframe, there will be a tuple in the list where the first
        value is the node_id and the second is an AgeDistribution object that can be used
        when creating a Demographics object.

        The "rate" column is assumed to be the fraction of people in that year and age range.
        The dataframe is also assumed to only have the data for one year.

        NOTE: EMOD expects the ReslutValues/Ages to be maximums of the bin. This implies that
        if the last age has a DistributionValue = 1.0, then there should be no people aged
        greater than this last age.  It also means that the first age is also a minimum.
        For example, if the first age were 1.0, then there can be zero people less than 1.0.
        !!!THIS APPLIES ONLY TO THE OUTPUT OF THIS FUNCTION AND NOT THE INPUT!!!
        """

        unique_min_year_list = self.df[YearAgeRate.COL_NAME_MIN_YEAR].unique()
        if len(unique_min_year_list) != 1:
            msg = f"To be converted to an AgeDistribution, the dataframe must have only one " \
                  f"'{YearAgeRate.COL_NAME_MIN_YEAR}'.\n"
            msg += f"The dataframe has the following unique values for 'min_year': {unique_min_year_list}"
            raise ValueError(msg)

        age_distributions = []

        node_group_collection = self.df.groupby(YearAgeRate.COL_NAME_NODE_ID)
        for node_id, node_group in node_group_collection:
            # -----------------------------------------------------------------------------
            # --- Turn the collection of individual fractions into cumulative distribution
            # -----------------------------------------------------------------------------
            fractions = node_group[YearAgeRate.COL_NAME_RATE]
            cumulative_fractions = []
            cum = 0
            total = sum(fractions)
            for frac in fractions:
                # ensure fractions are normalized
                cum += round(frac / total, 6)
                cumulative_fractions.append(cum)
            cumulative_fractions[-1] = 1.0  # ensure last value is exactly 1
            ages = node_group[YearAgeRate.COL_NAME_MIN_AGE].tolist()

            # ------------------------------------------------------------------------
            # --- Make Ages Maximums (see above)
            # --- Insert a zero at the beginning of the fractions to force the values
            # --- to be associated with next age or the maximum of the age range.
            # --- Insert 125 at the end of the ages so that the 1.0 of the fractions
            # --- is associated with 125 - no one is older than 125.
            # ------------------------------------------------------------------------
            cumulative_fractions.insert(0, 0.0)
            ages.append(125.0)

            # -------------------------------
            # --- Convert to dictionary/JSON
            # -------------------------------
            # TODO: we should be creating and returning distribution objects by constructor here, not json/converting
            #  https://github.com/InstituteforDiseaseModeling/emodpy-hiv-old/issues/316
            ages_and_fractions = {
                'ResultValues': ages,
                'DistributionValues': cumulative_fractions,
                'NumDistributionAxes': 0,
                'ResultUnits': 'years',
                'ResultScaleFactor': 365
            }

            # ----------------------------------------------------------
            # --- Convert from dictionary/JSON to AgeDistribution object
            # ----------------------------------------------------------
            age_distribution = AgeDistribution.from_dict(distribution_dict=ages_and_fractions)
            age_distributions.append((node_id, age_distribution))

        return age_distributions

    @staticmethod
    def _node_group_to_distribution_json(node_group: pd.DataFrame,
                                         stepwise_for_year: bool = True) -> Dict:
        """
        Since fertility and mortality data is formatted the same, this method converts the dataframe
        to an EMOD distribution dictionary.  This dictionary can be used with FertilityDistribution's
        or MortalityDistribution's from_dict() method to create the appropriate distribution object.

        The method assumes that the user wants the data in a step-wise format.  That is, for a
        calendar year range and age range, the user wants EMOD to produce the same value/rate
        for the entire range.  The rate doesn't change until the year or age moves to a new range.

        For the max age of the last bin, a value of 125 is used and, for the max_year of the last bin,
        a value of 2100 is used.
        """
        new_node_group = pd.DataFrame()

        # ---------------------------------------------------------------------------------
        # --- Break the collection for the node into groups by min_year
        # --- Then, find the max age of the bin based on the next age (in the next row)
        # --- Things should be sorted such that for the given min_year, the ages are unique
        # --- Create a new group that has ths new/temporary column
        # ---------------------------------------------------------------------------------
        min_year_group_collection = node_group.groupby(YearAgeRate.COL_NAME_MIN_YEAR)
        for min_year, min_year_group in min_year_group_collection:
            min_year_group["max_age"] = min_year_group[YearAgeRate.COL_NAME_MIN_AGE].shift(-1) - 0.001
            min_year_group["max_age"] = min_year_group["max_age"].fillna(125)
            new_node_group = pd.concat([new_node_group, min_year_group])
        new_node_group = new_node_group.sort_values(by=[YearAgeRate.COL_NAME_MIN_YEAR, YearAgeRate.COL_NAME_MIN_AGE],
                                                    ascending=True)

        # ----------------------------------------------------------------------------------
        # --- Duplicate the set of rows, but have the min_age be the max_age determined above
        # ----------------------------------------------------------------------------------
        node_group_max_age = new_node_group.copy()
        node_group_max_age.drop([YearAgeRate.COL_NAME_MIN_AGE], axis=1, inplace=True)
        node_group_max_age.rename({"max_age": YearAgeRate.COL_NAME_MIN_AGE}, axis=1, inplace=True)
        new_node_group.drop(["max_age"], axis=1, inplace=True)
        new_node_group = pd.concat([new_node_group, node_group_max_age]).reset_index()
        new_node_group = new_node_group.sort_values(by=[YearAgeRate.COL_NAME_MIN_YEAR, YearAgeRate.COL_NAME_MIN_AGE],
                                                    ascending=True)

        # ------------------------------------------------------------------------------------
        # --- Do like we did for min_age, but do it for min_year.  That is, create a new
        # --- max_year column that has the value of min_year in the next row minus a smidge.
        # ------------------------------------------------------------------------------------
        if stepwise_for_year:
            new_node_group2 = pd.DataFrame()
            min_age_group_collection = new_node_group.groupby(YearAgeRate.COL_NAME_MIN_AGE)
            for min_age, min_age_group in min_age_group_collection:
                min_age_group["max_year"] = min_age_group[YearAgeRate.COL_NAME_MIN_YEAR].shift(-1) - 0.001
                min_age_group["max_year"] = min_age_group["max_year"].fillna(2101)
                new_node_group2 = pd.concat([new_node_group2, min_age_group])
            new_node_group = new_node_group2.sort_values(by=[YearAgeRate.COL_NAME_MIN_YEAR, YearAgeRate.COL_NAME_MIN_AGE],
                                                         ascending=True)

            # ----------------------------------------------------------------------------------
            # --- Duplicate the set of rows, but have the min_year be the max_year determined above
            # ----------------------------------------------------------------------------------
            node_group_max_year = new_node_group.copy()
            node_group_max_year.drop([YearAgeRate.COL_NAME_MIN_YEAR], axis=1, inplace=True)
            node_group_max_year.rename({"max_year": YearAgeRate.COL_NAME_MIN_YEAR}, axis=1, inplace=True)
            new_node_group.drop(["max_year"], axis=1, inplace=True)
            new_node_group = pd.concat([new_node_group, node_group_max_year])
            new_node_group = new_node_group.sort_values(by=[YearAgeRate.COL_NAME_MIN_YEAR, YearAgeRate.COL_NAME_MIN_AGE],
                                                        ascending=True)
        else:
            # Not coming up with a quick way to determine that the difference between years is 5.
            difference_between_years = 5
            new_node_group[YearAgeRate.COL_NAME_MIN_YEAR] = new_node_group[YearAgeRate.COL_NAME_MIN_YEAR] + difference_between_years / 2.0

        # -----------------------------------------
        # --- Create the JSON for the distribution
        # -----------------------------------------
        dist_dict = {
            "NumDistributionAxes": 2,
            "AxisNames": ["age", "year"],
            "AxisUnits": ["years", "simulation_year"],
            "AxisScaleFactors": [365, 1],
            "PopulationGroups": [
                [],
                []
            ],
            "ResultScaleFactor": (1.0 / 365.0),  # 2.73972602739726e-03,
            "ResultUnits": "",
            "ResultValues": []
        }

        min_ages  = sorted(new_node_group[YearAgeRate.COL_NAME_MIN_AGE ].unique().tolist()) # Noqa: E221, E202
        min_years = sorted(new_node_group[YearAgeRate.COL_NAME_MIN_YEAR].unique().tolist())

        dist_dict["PopulationGroups"][0] = min_ages
        dist_dict["PopulationGroups"][1] = min_years

        min_age_group_collection = new_node_group.groupby(YearAgeRate.COL_NAME_MIN_AGE)
        for min_age, min_age_group in min_age_group_collection:
            rate_values = min_age_group[YearAgeRate.COL_NAME_RATE].tolist()
            dist_dict["ResultValues"].append(rate_values)

        return dist_dict

    def to_fertility_distributions(self) -> List[Tuple[int, FertilityDistribution]]:
        """
        Convert this YearAgeRate object to a list of tuples of node_id and FertilityDistribution.

        The method assumes that the user wants the data in a step-wise format.  That is, for a
        calendar year range and age range, the user wants EMOD to produce the same value/rate
        for the entire range.  The rate doesn't change until the year or age moves to a new range.

        For the max age of the last bin, a value of 125 is used and, for the max_year of the last bin,
        a value of 2100 is used.  Using these maximums still produces the correct values because we
        want the result constant for the entire range.  Having the range wider just produces
        the same constant.
        """
        fertility_distributions = []

        node_group_collection = self.df.groupby(YearAgeRate.COL_NAME_NODE_ID)
        for node_id, node_group in node_group_collection:
            # TODO: we should be creating and returning distribution objects in this call, not json/converting below
            #  https://github.com/InstituteforDiseaseModeling/emodpy-hiv-old/issues/316
            dist_dict = YearAgeRate._node_group_to_distribution_json(node_group)
            dist_dict["ResultUnits"] = "annual birth rate per 1000 women"
            dist_dict["ResultScaleFactor"] = dist_dict["ResultScaleFactor"] / 1000.0

            distribution = FertilityDistribution.from_dict(distribution_dict=dist_dict)
            fertility_distributions.append((node_id, distribution))

        return fertility_distributions

    def to_mortality_distributions(self, stepwise_for_year: bool = True) -> Dict[int, MortalityDistribution]:
        """
        Convert this YearAgeRate object to a dict of node_id: MortalityDistribution entries.

        The method assumes that the user wants the data in a step-wise format.  That is, for a
        calendar year range and age range, the user wants EMOD to produce the same value/rate
        for the entire range.  The rate doesn't change until the year or age moves to a new range.

        For the max age of the last bin, a value of 125 is used and, for the max_year of the last bin,
        a value of 2100 is used.  Using these maximums still produces the correct values because we
        want the result constant for the entire range.  Having the range wider just produces
        the same constant.

        Args:
            stepwise_for_year:
                If true, the age and calendar year both in step-wise format.  If false, calendar year
                is adjust by 2.5 and the linear interpolation will be used between calendar years.
        """
        mortality_distributions = {}

        node_group_collection = self.df.groupby(YearAgeRate.COL_NAME_NODE_ID)
        for node_id, node_group in node_group_collection:
            # TODO: we should be creating and returning distribution objects in this call, not json/converting below
            #  https://github.com/InstituteforDiseaseModeling/emodpy-hiv-old/issues/316
            dist_dict = YearAgeRate._node_group_to_distribution_json(node_group, stepwise_for_year=stepwise_for_year)
            dist_dict["ResultUnits"] = "annual death rate for an individual"

            distribution = MortalityDistribution.from_dict(distribution_dict=dist_dict)
            mortality_distributions[node_id] = distribution

        return mortality_distributions


def plot(year_age_rate_list: List[YearAgeRate],
         title: str = None,
         node_id: int = 0,
         img_dir: str = None,
         filename_to_save_to: str = None):
    """
    Create a plot window with one subplot for each min_age value and where the subplot
    will has min_year on the x-axis and "rate" on the y-axis.  Each YearAgeRate object
    in the 'year_age_rate_list' will have one curve on each subplot

    Args:
        year_age_rate_list: A list of YearAgeRate objects that have same min_age values
            for all min_year values.  The min_year values do not need to be the same,
            just the min_ages.

        title:
            The title of the plot window.

        node_id:
            Data will be extracted from the YearAgeRate objects for this node.

        img_dir: If this is defined, the images are saved to this directory.  If not defined,
            the images are displayed in a window.

        filename_to_save_to: The name of the file to save the image to.  This is only used if
            img_dir is defined.
    """
    from math import sqrt, ceil
    import matplotlib.pyplot as plt

    def getColor(idx):
        colors = ['r', 'b', 'g', 'c', 'm', 'y', 'k']
        return colors[idx % len(colors)]

    def get_line_style(idx):
        ls_list = ["-", "--", ":", "-."]
        return ls_list[idx % len(ls_list)]

    if img_dir:
        plt.figure(figsize=(15, 10))

    min_age_list = year_age_rate_list[0].df[YearAgeRate.COL_NAME_MIN_AGE].unique().tolist()
    num_sub_plots_total = len(min_age_list)
    num_sub_plots_x = ceil(sqrt(num_sub_plots_total))
    num_sub_plots_y = ceil(float(num_sub_plots_total) / float(num_sub_plots_x))

    i_sub_plot = 0
    for min_age in min_age_list:
        sub_plot_ix = i_sub_plot % num_sub_plots_x
        sub_plot_iy = int(i_sub_plot / num_sub_plots_x)
        sub_plot = plt.subplot2grid((num_sub_plots_y, num_sub_plots_x), (sub_plot_iy, sub_plot_ix))

        i_yar = 0
        for yar in year_age_rate_list:
            node_df = yar.df[yar.df[YearAgeRate.COL_NAME_NODE_ID] == node_id]
            min_age_df = node_df[node_df[YearAgeRate.COL_NAME_MIN_AGE] == min_age]

            x = min_age_df[YearAgeRate.COL_NAME_MIN_YEAR].tolist()
            y = min_age_df[YearAgeRate.COL_NAME_RATE    ].tolist() # Noqa: E202

            color = getColor(i_yar)
            ls = get_line_style(i_yar)
            sub_plot.plot(x, y, color, linewidth=2, linestyle=ls)
            plt.setp(sub_plot.get_xticklabels(), fontsize='7')
            plt.title(str(min_age), fontsize='9')

            i_yar += 1
        i_sub_plot += 1

    if title is None:
        title = "Data for Node=" + str(node_id)
    plt.suptitle(title)

    plt.subplots_adjust(left=0.04, right=0.99, bottom=0.02, top=0.9, wspace=0.3, hspace=0.3)

    if img_dir:
        import os
        if not os.path.exists(img_dir):
            os.makedirs(img_dir)
        if filename_to_save_to is None:
            filename_to_save_to = "YearAgeRate.png"
        fn = os.path.join(img_dir, filename_to_save_to)
        print(fn)
        plt.savefig(fn, dpi=300)
    else:
        plt.show()
    plt.close()



from emodpy.utils.distributions import BaseDistribution
from emodpy.utils.distributions import ConstantDistribution
from emodpy.utils.distributions import ExponentialDistribution
from emodpy.utils.distributions import UniformDistribution
from emodpy.utils.distributions import GaussianDistribution
from emodpy.utils.distributions import PoissonDistribution
from emodpy.utils.distributions import LogNormalDistribution
from emodpy.utils.distributions import WeibullDistribution
from emodpy.utils.distributions import DualConstantDistribution
from emodpy.utils.distributions import DualExponentialDistribution
from emodpy.utils.distributions import BimodalDistribution


# __all_exports: A list of classes that are intended to be exported from this module.
__all_exports = [
    BaseDistribution,
    ConstantDistribution,
    ExponentialDistribution,
    UniformDistribution,
    GaussianDistribution,
    PoissonDistribution,
    LogNormalDistribution,
    WeibullDistribution,
    DualConstantDistribution,
    DualExponentialDistribution,
    BimodalDistribution,
]

# The following loop sets the __module__ attribute of each class in __all_exports to the name of the current module.
# This is done to ensure that when these classes are imported from this module, their __module__ attribute correctly
# reflects their source module.

for _ in __all_exports:
    _.__module__ = __name__

# __all__: A list that defines the public interface of this module.
# This is essential to ensure that Sphinx builds documentation for these classes, including those that are imported
# from emodpy.
# It contains the names of all the classes that should be accessible when this module is imported using the syntax
# 'from module import *'.
# Here, it is set to the names of all classes in __all_exports.

__all__ = [_.__name__ for _ in __all_exports]



from typing import Dict

from emod_api.demographics.updateable import Updateable


class CondomUsageParameters(Updateable):
    def __init__(self,
                 min: float = 0,
                 mid: float = 0,
                 max: float = 0,
                 rate: float = 0):
        """
            Parameterizes condom usage over time, a component of EMOD relationship parameters.

            https://docs.idmod.org/projects/emod-hiv/en/latest/sti-model-relationships.html#condom-usage
            https://docs.idmod.org/projects/emod-hiv/en/latest/parameter-demographics.html#relationship-parameters

            Args:
                min: minimum condom usage probability (pre-inflection point)
                mid: inflection point in condom usage (a year)
                max: maximum condom usage probability (post-inflection point)
                rate: slope of condom usage at inflection point
        """
        super().__init__()
        self.min = min
        self.mid = mid
        self.max = max
        self.rate = rate

    def to_dict(self) -> Dict:
        usage = {
            'Min': self.min,
            'Mid': self.mid,
            'Max': self.max,
            'Rate': self.rate
        }
        return usage

    @classmethod
    def from_dict(cls, d: Dict) -> '__class__':
        return cls(min=d['Min'], mid=d['Mid'], max=d['Max'], rate=d['Rate'])
